# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Mechanical Collection Classes."""

from __future__ import annotations

from collections import defaultdict
from copy import copy
from dataclasses import dataclass, field
from functools import reduce
from typing import Any, ClassVar, Dict, List, Optional, Union

from PHX.model import hvac
from PHX.model.enums.hvac import DeviceType, PhxSupportiveDeviceType
from PHX.model.hvac.heat_pumps import AnyPhxHeatPump
from PHX.model.hvac.heating import AnyPhxHeater
from PHX.model.hvac.piping import PhxRecirculationParameters
from PHX.model.hvac.renewable_devices import AnyRenewableDevice, PhxDevicePhotovoltaic
from PHX.model.hvac.supportive_devices import PhxSupportiveDevice
from PHX.model.hvac.ventilation import (
    AnyPhxExhaustVent,
    AnyPhxVentilation,
    PhxExhaustVentilatorDryer,
    PhxExhaustVentilatorRangeHood,
    PhxExhaustVentilatorUserDefined,
)
from PHX.model.hvac.water import AnyWaterTank

# ------------------------------------------------------------------------------


class NoDeviceFoundError(Exception):
    def __init__(self, _id_num):
        self.msg = f"Error: Cannot locate the Mechanical Device with id num: {_id_num}"
        super().__init__(self.msg)


class NoSupportiveDeviceUnitFoundError(Exception):
    def __init__(self, _id_num):
        self.msg = f"Error: Cannot locate the Supportive Device with id num: {_id_num}"
        super().__init__(self.msg)


class NoRenewableDeviceUnitFoundError(Exception):
    def __init__(self, _id_num):
        self.msg = f"Error: Cannot locate the Renewable Device with id num: {_id_num}"
        super().__init__(self.msg)


# ------------------------------------------------------------------------------


@dataclass
class PhxZoneCoverage:
    """Percentage of the building load-type covered by the subsystem."""

    zone_num: int = 1
    heating: float = 1.0
    cooling: float = 1.0
    ventilation: float = 1.0
    humidification: float = 1.0
    dehumidification: float = 1.0


AnyMechDevice = Union[AnyPhxVentilation, AnyPhxHeater, AnyPhxHeatPump, AnyWaterTank, AnyRenewableDevice]


# ------------------------------------------------------------------------------


@dataclass
class PhxRenewableDeviceCollection:
    """A Collection of PHX Renewable Energy Devices."""

    _count: ClassVar[int] = 0

    id_num: int = field(init=False, default=0)
    display_name: str = "Renewable Energy Device Collection"
    _devices: Dict[str, AnyRenewableDevice] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def devices(self) -> List[AnyRenewableDevice]:
        return list(self._devices.values())

    @property
    def pv_devices(self) -> List[PhxDevicePhotovoltaic]:
        """Return a List of all the Photovoltaic devices"""
        return [d for d in self.devices if d.device_type == DeviceType.PHOTOVOLTAIC]

    def clear_all_devices(self) -> None:
        """Reset the collection to an empty dictionary."""
        self._devices = {}

    def device_in_collection(self, _device_key) -> bool:
        """Return True if a PHX Renewable Device with the matching key is in the collection."""
        return _device_key in self._devices.keys()

    def get_device_by_key(self, _key: str) -> Optional[AnyRenewableDevice]:
        """Returns the PHX Renewable Device with the matching key, or None if not found.

        Arguments:
        ----------
            * _key (str): The key to search the collection for.

        Returns:
        --------
            * (Optional[PhxRenewableDevice]) The supportive device with
                the matching key, or None if not found.
        """
        return self._devices.get(_key, None)

    def get_device_by_id(self, _id_num: int) -> AnyRenewableDevice:
        """Returns a PHX Renewable Device from the collection which has a matching id-num.

        Arguments:
        ----------
            * _id_num (int): The Renewable Device id-number to search for.

        Returns:
        --------
            * (AnyRenewableDevice): The Renewable Device found with the
                matching ID-Number. Or Error if not found.
        """
        for device in self._devices.values():
            if device.id_num == _id_num:
                return device

        raise NoRenewableDeviceUnitFoundError(_id_num)

    def add_new_device(self, _key: str, _d: AnyRenewableDevice) -> None:
        """Adds a new PHX Supportive Device to the collection.

        Arguments:
        ----------
            * _key (str): The key to use when storing the new device
            * _device (AnyRenewableDevice): The new PHX Renewable Device
                to add to the collection.

        Returns:
        --------
            * None
        """
        if not _d:
            return
        self._devices[_key] = _d

    def group_devices_by_identifier(self, _devices: List[AnyRenewableDevice]) -> Dict[str, List[AnyRenewableDevice]]:
        d = defaultdict(list)
        for device in _devices:
            d[device.identifier].append(copy(device))
        return d

    def merge_group_of_devices(self, _groups: Dict[str, List[AnyRenewableDevice]]) -> List[AnyRenewableDevice]:
        """Merge a group of Dict of device-groups together into a single device."""
        merged_devices = []
        for group in _groups.values():
            merged_devices.append(reduce(lambda d1, d2: d1 + d2, group))
        return merged_devices

    def merge_all_devices(self) -> None:
        """Merge all the devices in the collection together by identifier."""
        pv_devices = self.group_devices_by_identifier(self.pv_devices)

        # -- start fresh
        self.clear_all_devices()

        # -- Merge, and add back the merged devices
        for device in self.merge_group_of_devices(pv_devices):
            self.add_new_device(device.identifier, device)

    def __iter__(self):
        """Get each device in the PhxRenewableDeviceCollection, one at a time."""
        for _ in self.devices:
            yield _

    def __len__(self) -> int:
        """Number of devices in the PhxRenewableDeviceCollection"""
        return len(self.devices)

    def __bool__(self) -> bool:
        return bool(self._devices)


@dataclass
class PhxSupportiveDeviceCollection:
    """A Collection of PHX Supportive Devices."""

    _count: ClassVar[int] = 0

    id_num: int = field(init=False, default=0)
    display_name: str = "Supportive Device Collection"
    _devices: Dict[str, PhxSupportiveDevice] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def devices(self) -> List[PhxSupportiveDevice]:
        return list(self._devices.values())

    @property
    def heat_circulating_pumps(self) -> List[PhxSupportiveDevice]:
        """Return a List of all the Kitchen Hood devices"""
        return [d for d in self.devices if d.device_type == PhxSupportiveDeviceType.HEAT_CIRCULATING_PUMP]

    @property
    def dhw_circulating_pumps(self) -> List[PhxSupportiveDevice]:
        """Return a List of all the Kitchen Hood devices"""
        return [d for d in self.devices if d.device_type == PhxSupportiveDeviceType.DHW_CIRCULATING_PUMP]

    @property
    def dhw_storage_pumps(self) -> List[PhxSupportiveDevice]:
        """Return a List of all the Kitchen Hood devices"""
        return [d for d in self.devices if d.device_type == PhxSupportiveDeviceType.DHW_STORAGE_LOAD_PUMP]

    @property
    def other_devices(self) -> List[PhxSupportiveDevice]:
        """Return a List of all the Kitchen Hood devices"""
        return [d for d in self.devices if d.device_type == PhxSupportiveDeviceType.OTHER]

    def clear_all_devices(self) -> None:
        """Reset the collection to an empty dictionary."""
        self._devices = {}

    def device_in_collection(self, _device_key) -> bool:
        """Return True if a PHX Supportive Device with the matching key is in the collection."""
        return _device_key in self._devices.keys()

    def get_device_by_key(self, _key: str) -> Optional[PhxSupportiveDevice]:
        """Returns the PHX Supportive Device with the matching key, or None if not found.

        Arguments:
        ----------
            * _key (str): The key to search the collection for.

        Returns:
        --------
            * (Optional[PhxSupportiveDevice]) The supportive device with
                the matching key, or None if not found.
        """
        return self._devices.get(_key, None)

    def get_device_by_id(self, _id_num: int) -> PhxSupportiveDevice:
        """Returns a PHX Supportive Device from the collection which has a matching id-num.

        Arguments:
        ----------
            * _id_num (int): The Supportive Device id-number to search for.

        Returns:
        --------
            * (PhxSupportiveDevice): The Supportive Device found with the
                matching ID-Number. Or Error if not found.
        """
        for device in self._devices.values():
            if device.id_num == _id_num:
                return device

        raise NoSupportiveDeviceUnitFoundError(_id_num)

    def add_new_device(self, _key: str, _d: PhxSupportiveDevice) -> None:
        """Adds a new PHX Supportive Device to the collection.

        Arguments:
        ----------
            * _key (str): The key to use when storing the new device
            * _device (hvac.ventilation.AnyPhxExhaustVent): The new PHX Supportive Device
                to add to the collection.

        Returns:
        --------
            * None
        """
        if not _d:
            return
        self._devices[_key] = _d

    def group_devices_by_identifier(self, _devices: List[PhxSupportiveDevice]) -> Dict[str, List[PhxSupportiveDevice]]:
        d = defaultdict(list)
        for device in _devices:
            d[device.identifier].append(copy(device))
        return d

    def merge_group_of_devices(self, _groups: Dict[str, List[PhxSupportiveDevice]]) -> List[PhxSupportiveDevice]:
        """Merge a group of Dict of device-groups together into a single device."""
        merged_devices = []
        for group in _groups.values():
            merged_devices.append(reduce(lambda d1, d2: d1 + d2, group))
        return merged_devices

    def merge_all_devices(self) -> None:
        """Merge all the devices in the collection together by identifier."""
        heat_circulating_pumps = self.group_devices_by_identifier(self.heat_circulating_pumps)
        dhw_circulating_pumps = self.group_devices_by_identifier(self.dhw_circulating_pumps)
        dhw_storage_pumps = self.group_devices_by_identifier(self.dhw_storage_pumps)
        others = self.group_devices_by_identifier(self.other_devices)

        # -- start fresh
        self.clear_all_devices()

        # -- Merge, and add back the merged devices
        for device in self.merge_group_of_devices(heat_circulating_pumps):
            self.add_new_device(device.identifier, device)
        for device in self.merge_group_of_devices(dhw_circulating_pumps):
            self.add_new_device(device.identifier, device)
        for device in self.merge_group_of_devices(dhw_storage_pumps):
            self.add_new_device(device.identifier, device)
        for device in self.merge_group_of_devices(others):
            self.add_new_device(device.identifier, device)

    def __iter__(self):
        """Get each device in the PhxSupportiveDeviceCollection, one at a time."""
        for _ in self.devices:
            yield _

    def __len__(self) -> int:
        """Number of devices in the PhxSupportiveDeviceCollection"""
        return len(self.devices)

    def __bool__(self) -> bool:
        return bool(self._devices)


@dataclass
class PhxExhaustVentilatorCollection:
    """A Collection of PHX Exhaust Ventilation Devices."""

    _count: ClassVar[int] = 0

    id_num: int = field(init=False, default=0)
    display_name: str = "Exhaust Ventilator Collection"
    _devices: Dict[str, AnyPhxExhaustVent] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def devices(self) -> List[AnyPhxExhaustVent]:
        return list(self._devices.values())

    @property
    def kitchen_hood_devices(self) -> List[PhxExhaustVentilatorRangeHood]:
        """Return a List of all the Kitchen Hood devices"""
        return [d for d in self.devices if isinstance(d, PhxExhaustVentilatorRangeHood)]

    @property
    def dryer_devices(self) -> List[PhxExhaustVentilatorDryer]:
        """Return a List of all the Dryer devices"""
        return [d for d in self.devices if isinstance(d, PhxExhaustVentilatorDryer)]

    @property
    def user_determined_devices(self) -> List[PhxExhaustVentilatorUserDefined]:
        """Return a List of all the User-Determined devices"""
        return [d for d in self.devices if isinstance(d, PhxExhaustVentilatorUserDefined)]

    def clear_all_devices(self) -> None:
        """Reset the collection to an empty dictionary."""
        self._devices = {}

    def device_in_collection(self, _device_key) -> bool:
        """Return True if a PHX Exhaust Ventilator with the matching key is in the collection."""
        return _device_key in self._devices.keys()

    def get_ventilator_by_key(self, _key: str) -> Optional[AnyPhxExhaustVent]:
        """Returns the PHX Exhaust Ventilator with the matching key, or None if not found.

        Arguments:
        ----------
            * _key (str): The key to search the collection for.

        Returns:
        --------
            * (Optional[hvac.ventilation.AnyPhxExhaustVent]) The Mechanical device with
                the matching key, or None if not found.
        """
        return self._devices.get(_key, None)

    def get_ventilator_by_id(self, _id_num: int) -> AnyPhxExhaustVent:
        """Returns a PHX Exhaust Ventilator from the collection which has a matching id-num.

        Arguments:
        ----------
            * _id_num (int): The Exhaust Ventilator id-number to search for.

        Returns:
        --------
            * (hvac.ventilation.AnyPhxExhaustVent): The Exhaust Ventilator found with the
                matching ID-Number. Or Error if not found.
        """
        for device in self._devices.values():
            if device.id_num == _id_num:
                return device

        raise NoDeviceFoundError(_id_num)

    def add_new_ventilator(self, _key: str, _d: AnyPhxExhaustVent) -> None:
        """Adds a new PHX Exhaust Ventilator to the collection.

        Arguments:
        ----------
            * _key (str): The key to use when storing the new mechanical device
            * _device (hvac.ventilation.AnyPhxExhaustVent): The new PHX exhaust ventilator
                to add to the collection.

        Returns:
        --------
            * None
        """
        if not _d:
            return
        self._devices[_key] = _d

    def merge_all_devices(self):
        """Merge all the devices in the collection together by type."""
        # -- Keep a copy of the devices the collection started with
        kitchen_hood_devices = copy(self.kitchen_hood_devices)
        dryer_devices = copy(self.dryer_devices)
        user_determined_devices = copy(self.user_determined_devices)

        # -- start fresh
        self.clear_all_devices()

        # -- Add back the merged devices
        if kitchen_hood_devices:
            kitchen_hood = reduce(lambda d1, d2: d1 + d2, kitchen_hood_devices)
            self.add_new_ventilator(kitchen_hood.identifier, kitchen_hood)

        if dryer_devices:
            dryer = reduce(lambda d1, d2: d1 + d2, dryer_devices)
            self.add_new_ventilator(dryer.identifier, dryer)

        if user_determined_devices:
            ud_devices = reduce(lambda d1, d2: d1 + d2, user_determined_devices)
            self.add_new_ventilator(ud_devices.identifier, ud_devices)

    def __iter__(self):
        """Get each device in the PhxExhaustVentilatorCollection, one at a time."""
        for _ in self.devices:
            yield _

    def __len__(self) -> int:
        """Number of devices in the PhxExhaustVentilatorCollection"""
        return len(self.devices)

    def __bool__(self) -> bool:
        return bool(self._devices)


# ------------------------------------------------------------------------------


@dataclass
class PhxMechanicalSystemCollection:
    """A Collection of all the mechanical devices (heating, cooling, etc) and distribution in the project"""

    _count: ClassVar[int] = 0

    id_num: int = field(init=False, default=0)
    display_name: str = "Ideal Air System"
    sys_type_num: int = 1
    sys_type_str: str = "User defined (ideal system)"
    zone_coverage: PhxZoneCoverage = field(default_factory=PhxZoneCoverage)

    _devices: Dict[str, AnyMechDevice] = field(default_factory=dict)
    _distribution_hw_recirculation_params: PhxRecirculationParameters = field(
        default_factory=PhxRecirculationParameters
    )
    _distribution_piping_trunks: Dict[str, hvac.PhxPipeTrunk] = field(default_factory=dict)
    _distribution_piping_recirc: Dict[str, hvac.PhxPipeElement] = field(default_factory=dict)
    _distribution_num_hw_tap_points: int = 1
    _distribution_ducting: Dict[str, hvac.PhxDuctElement] = field(default_factory=dict)

    supportive_devices: PhxSupportiveDeviceCollection = field(default_factory=PhxSupportiveDeviceCollection)
    renewable_devices: PhxRenewableDeviceCollection = field(default_factory=PhxRenewableDeviceCollection)

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    # -------------------------------------------------------------------------
    #  -- Mechanical Devices

    @property
    def devices(self) -> List[AnyMechDevice]:
        """Returns a list of the 'devices' in the collection."""
        return list(sorted(self._devices.values(), key=lambda d: d.display_name))

    @property
    def ventilation_devices(self) -> List[hvac.AnyPhxVentilation]:
        """Returns a list of the 'Ventilation' devices in the collection."""
        return [_ for _ in self.devices if isinstance(_, hvac.PhxDeviceVentilation)]

    @property
    def space_heating_devices(self) -> List[hvac.AnyPhxHeater]:
        """Returns a list of the 'Space Heating' devices in the collection."""
        return [_ for _ in self.devices if isinstance(_, hvac.PhxHeatingDevice)]

    @property
    def heat_pump_devices(self) -> List[hvac.PhxHeatPumpDevice]:
        """Returns a list of all the Heat-Pump devices in the collection."""
        return [d for d in self.devices if isinstance(d, hvac.PhxHeatPumpDevice)]

    @property
    def cooling_devices(self) -> List[hvac.PhxHeatPumpDevice]:
        """Returns a list of all the Cooling devices (heat pumps) in the collection."""
        return [d for d in self.heat_pump_devices if d.usage_profile.cooling == True]

    def device_in_collection(self, _device_key) -> bool:
        """Return True if the a Mech device with the matching key is already in the collection."""
        return _device_key in self._devices.keys()

    def get_mech_device_by_key(self, _key: str) -> Optional[hvac.PhxMechanicalDevice]:
        """Returns the mechanical device with the matching key, or None if not found.

        Arguments:
        ----------
            * _key (str): The key to search the collection for.

        Returns:
        --------
            * (Optional[hvac.PhxMechanicalDevice]) The Mechanical device with
                the matching key, or None if not found.
        """
        return self._devices.get(_key, None)

    def get_mech_device_by_id(self, _id_num: int) -> hvac.PhxMechanicalDevice:
        """Returns a Mechanical Device from the collection which has a matching id-num.

        Arguments:
        ----------
            * _id_num (int): The Mechanical Device id-number to search for.

        Returns:
        --------
            * (hvac.PhxMechanicalDevice): The Mechanical Device found with the
                matching ID-Number. Or Error if not found.
        """
        for device in self._devices.values():
            if device.id_num == _id_num:
                return device

        raise NoDeviceFoundError(_id_num)

    def add_new_mech_device(self, _key: str, _d: AnyMechDevice) -> None:
        """Adds a new PHX Mechanical device to the collection.

        Arguments:
        ----------
            * _key (str): The key to use when storing the new mechanical device
            * _device (_base.PhxMechanicalDevice): The new PHX mechanical device to
                add to the collection.

        Returns:
        --------
            * None
        """
        self._devices[_key] = _d

    # -------------------------------------------------------------------------
    #  -- Distribution Piping

    def add_distribution_piping(self, _p: hvac.PhxPipeTrunk) -> None:
        """Add a new DHW Trunc-Pipe to the System.
        Note that all piping must be in a: Trunc > Branch > Fixture config.
        """
        self._distribution_piping_trunks[_p.identifier] = _p

    def add_recirc_piping(self, _p: hvac.PhxPipeElement) -> None:
        """Add a new DHW Recirc-Pipe to the System."""
        self._distribution_piping_recirc[_p.identifier] = _p

    @property
    def dhw_heating_devices(self) -> List[hvac.AnyPhxHeater]:
        """Returns a list of only the 'DHW Heating' devices (no tanks) in the collection."""
        return [
            _
            for _ in self.devices
            if isinstance(_, hvac.PhxHeatingDevice)
            and _.usage_profile.dhw_heating
            and _.device_type != DeviceType.WATER_STORAGE
        ]

    @property
    def dhw_tank_devices(self) -> List[hvac.AnyWaterTank]:
        """Returns a list of only the 'DHW Storage Tank' devices (no heaters) in the collection."""
        return [
            _
            for _ in self.devices
            if isinstance(_, hvac.PhxHotWaterTank)
            and _.usage_profile.dhw_heating
            and _.device_type == DeviceType.WATER_STORAGE
        ]

    @property
    def dhw_distribution_trunks(self) -> list[hvac.PhxPipeTrunk]:
        return sorted(self._distribution_piping_trunks.values(), key=lambda p: p.display_name)

    @property
    def dhw_distribution_piping(self) -> List[hvac.PhxPipeElement]:
        """Returns a list of ALL the DHW Piping from ALL Trunks, Branches, Twigs in the collection."""
        piping: List[hvac.PhxPipeElement] = []

        for trunk in self._distribution_piping_trunks.values():
            piping.append(trunk.pipe_element)
            for branch in trunk.branches:
                piping.append(branch.pipe_element)
                for fixture in branch.fixtures:
                    piping.append(fixture)

        return piping

    @property
    def dhw_distribution_piping_segments(self) -> List[hvac.PhxPipeSegment]:
        """Returns a list of ALL the DHW Piping Segments from ALL Trunks, Branches, Twigs in the collection."""
        pipe_segments: List[hvac.PhxPipeSegment] = []

        for trunk in self._distribution_piping_trunks.values():
            pipe_segments.extend(trunk.pipe_element.segments)
            for branch in trunk.branches:
                pipe_segments.extend(branch.pipe_element.segments)
                for fixture in branch.fixtures:
                    pipe_segments.extend(fixture.segments)

        return pipe_segments

    @property
    def dhw_distribution_piping_segments_by_diam(self) -> List[List[hvac.PhxPipeSegment]]:
        """Returns a list of the DHW branch-piping segments, grouped by diameter."""
        # -- Group piping segments by diameter
        d: Dict[float, List[hvac.PhxPipeSegment]] = defaultdict(list)
        for pipe_segment in self.dhw_distribution_piping_segments:
            d[pipe_segment.diameter_m].append(pipe_segment)

        return list(d.values())

    @property
    def dhw_recirc_piping(self) -> List[hvac.PhxPipeElement]:
        """Returns a list of all the DHW recirculation-piping in the collection."""
        return list(self._distribution_piping_recirc.values())

    @property
    def dhw_recirc_piping_segments_by_diam(self) -> List[List[hvac.PhxPipeSegment]]:
        """Returns a list of the DHW recirculation-piping segments, grouped by diameter."""
        # -- Group piping segments by diameter
        d: Dict[float, List[hvac.PhxPipeSegment]] = defaultdict(list)
        for pipe in self.dhw_recirc_piping:
            for segment in pipe.segments:
                d[segment.diameter_m].append(segment)

        return list(d.values())

    @property
    def dhw_recirc_total_length_m(self) -> float:
        return sum(_.length_m for _ in self.dhw_recirc_piping)

    @property
    def dhw_recirc_weighted_heat_loss_coeff(self) -> Optional[float]:
        """Return a length-weighted average pipe heat-loss coefficient."""
        if not self.dhw_recirc_total_length_m:
            return None

        weighted_total = 0.0
        for phx_pipe_element in self.dhw_recirc_piping:
            weighted_total += phx_pipe_element.weighted_pipe_heat_loss_coefficient * phx_pipe_element.length_m
        return weighted_total / self.dhw_recirc_total_length_m

    @property
    def dhw_distribution_total_length_m(self) -> float:
        return sum(_.length_m for _ in self.dhw_distribution_piping_segments)

    @property
    def dhw_distribution_weighted_diameter_mm(self) -> Optional[float]:
        """Return a length-weighted average diameter for the trunk/branch/twig piping."""
        if not self.dhw_distribution_total_length_m:
            return None

        weighted_total = 0.0
        for phx_pipe_element in self.dhw_distribution_piping:
            weighted_total += phx_pipe_element.weighted_diameter_mm * phx_pipe_element.length_m
        return weighted_total / self.dhw_distribution_total_length_m

    # -------------------------------------------------------------------------
    #  -- Distribution Ducting

    def add_vent_ducting(self, _d: hvac.PhxDuctElement) -> None:
        self._distribution_ducting[_d.identifier] = _d

    @property
    def vent_ducting(self) -> List[hvac.PhxDuctElement]:
        """Returns a list of all the Vent. Ducting in the collection."""
        return list(self._distribution_ducting.values())



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Occupancy (People) Utilization Schedule."""

from __future__ import annotations

import uuid
from dataclasses import dataclass, field
from typing import Any, ClassVar, Union


@dataclass
class PhxScheduleOccupancy:
    """A PHX Schedule for the Occupancy (People)."""

    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)
    identifier: Union[uuid.UUID, str] = field(default_factory=uuid.uuid4)
    display_name: str = "__unnamed_occupancy_schedule__"
    start_hour: float = 0.0
    end_hour: float = 1.0
    annual_utilization_days: float = 0.0
    relative_utilization_factor: float = 0.0  # Relative to the 'annual_utilization_days'

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    def __eq__(self, other: PhxScheduleOccupancy) -> bool:
        TOLERANCE = 0.001
        if self.display_name != other.display_name:
            return False
        if abs(self.start_hour - other.start_hour) > TOLERANCE:
            return False
        if abs(self.end_hour - other.end_hour) > TOLERANCE:
            return False
        if abs(self.annual_utilization_days - other.annual_utilization_days) > TOLERANCE:
            return False
        if abs(self.relative_utilization_factor - other.relative_utilization_factor) > TOLERANCE:
            return False
        if self.unique_key != other.unique_key:
            return False
        return True

    @property
    def annual_utilization_factor(self) -> float:
        """Return the annual Utilization Rate (0-1) relative to the entire year (8760 hours)"""

        operating_period_utilization_factor = self.annual_operating_hours / 8760  # Hrs / year

        return operating_period_utilization_factor * self.relative_utilization_factor

    @annual_utilization_factor.setter
    def annual_utilization_factor(self, _relative_utilization_factor: float):
        """Set the Relative Utilization Factor."""

        if _relative_utilization_factor is None:
            return

        # -- Re-Set the relative utilization factor to match
        self.start_hour = 0
        self.end_hour = 24
        self.annual_utilization_days = 365
        self.relative_utilization_factor = _relative_utilization_factor

    @property
    def daily_operating_hours(self) -> float:
        """Return the total Daily Operating Hours (end-hour - start-hour)."""
        return abs(self.end_hour - self.start_hour)

    @property
    def annual_operating_hours(self):
        """Return the total Annual Operating Hours (daily-hours * annual-util-days)."""
        return self.annual_utilization_days * self.daily_operating_hours

    @property
    def unique_key(self):
        """Return a key unique to this 'type' (collection of values) of pattern"""
        return f"{self.start_hour :.3f}_{self.end_hour :.3f}_{self.annual_utilization_days :.3f}_{self.annual_utilization_factor :.3f}_"

    @classmethod
    def constant_operation(cls) -> PhxScheduleOccupancy:
        """Return a constant operation (24/7, 365d) schedule."""
        return cls(
            display_name="Constant Operation",
            start_hour=0.0,
            end_hour=24.0,
            annual_utilization_days=365.0,
            relative_utilization_factor=1.0,
        )



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Controller Class for the PHPP "U-Values" worksheet."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Generator, List, Optional

from ph_units.unit_type import Unit

from PHX.model.constructions import PhxConstructionOpaque
from PHX.PHPP.phpp_localization import shape_model
from PHX.PHPP.phpp_model import uvalues_constructor
from PHX.PHPP.sheet_io.io_variants import VariantAssemblyLayerName
from PHX.xl import xl_app, xl_data
from PHX.xl.xl_data import col_offset


class NoEmptyConstructorError(Exception):
    def __init__(self):
        """Raises if there are no empty COnstructors in the worksheet."""
        self.msg = "No empty constructors found. Please remove some constructors and try again."
        super().__init__(self.msg)


@dataclass
class ExistingAssemblyData:
    name: str
    u_value: Unit
    r_value: Unit
    ext_exposure: str
    int_exposure: str


class UValues:
    """IO Controller for the PHPP "U-Values" worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.UValues) -> None:
        self.xl = _xl
        self.shape = _shape
        self._constructor_start_rows: List[int] = []
        self.cache = {}

    # -------------------------------------------------------------------------
    # -- Getters

    @property
    def all_constructor_start_rows(self) -> List[int]:
        """Return a list of all of the PHPP Constructor starting rows"""
        if not self._constructor_start_rows:
            self._constructor_start_rows = self.get_start_rows()
        return self._constructor_start_rows

    @property
    def used_constructor_start_rows(self) -> Generator[int, None, None]:
        """Return the PHPP Constructors that have any data in them, one at a time."""
        for row_num in self.all_constructor_start_rows:
            name_col = self.shape.constructor.inputs.display_name.column
            name_range = f"{name_col}{row_num + self.shape.constructor.inputs.name_row_offset}"
            if self.xl.get_data(self.shape.name, name_range):
                yield row_num

    def get_start_rows(self, _row_start: int = 1, _row_end: int = 1730) -> List[int]:
        """Reads through the U-Values worksheet and finds each of the constructor 'start' (title) rows.

        Arguments:
        ----------
            * _row_start: (int) default=1
            * _row_end: (int) default=1730

        Returns:
        -------
            * (List[int]): A list of all of the starting title row numbers found.
        """

        # -- Get the data from Excel in one operation
        col_data = self.xl.get_multiple_column_data(
            _sheet_name=self.shape.name,
            _col_start=self.shape.constructor.locator_col_header,
            _col_end=col_offset(self.shape.constructor.locator_col_header, 1),
            _row_start=_row_start,
            _row_end=_row_end,
        )

        # -- Find the starting row numbers
        constructors: List[int] = []
        for i, column_val in enumerate(col_data):
            if column_val[0] == self.shape.constructor.locator_string_header:
                constructors.append(i)
            elif column_val[0] == "Bauteil Nr.":  # Fuck you PHPP
                constructors.append(i)

        return constructors

    def get_constructor_phpp_id_by_name(
        self,
        _name: str,
        _row_start: int = 1,
        _row_end: int = 1730,
        _use_cache: bool = False,
    ) -> Optional[str]:
        """Returns the full PHPP-style value for the constructor with a specified name.

        ie: "Exterior Wall" in constructor 1 will return "01ud-Exterior Wall"

        Argument:
        ---------
            * _name: (str) The name to search for.
            * _row_start: (int) default=1
            * _row_end: (int) default=1730

        Returns:
        --------
            * (Optional[str]): The full PHPP-style id for the construction. ie: "01ud-MyConstruction"
        """

        if _use_cache:
            try:
                return self.cache[_name]
            except KeyError:
                pass

        row = self.xl.get_row_num_of_value_in_column(
            sheet_name=self.shape.name,
            row_start=_row_start,
            row_end=_row_end,
            col=str(self.shape.constructor.inputs.display_name.column),
            find=_name,
        )

        if not row:
            return

        id_num_offset = self.shape.constructor.inputs.phpp_id_num_col_offset
        if_num_column = col_offset(str(self.shape.constructor.inputs.display_name.column), id_num_offset)
        prefix = self.xl.get_data(self.shape.name, f"{if_num_column}{row}")
        name_with_id = f"{prefix}-{_name}"
        self.cache[_name] = name_with_id

        return name_with_id

    def get_used_constructor_names(self) -> List[str]:
        """Return a list of the used construction names."""

        name_col = self.shape.constructor.inputs.display_name.column

        construction_names = []
        for row_num in self.all_constructor_start_rows:
            assembly_name = self.xl.get_data(
                _sheet_name=self.shape.name,
                _range=f"{name_col}{row_num + self.shape.constructor.inputs.name_row_offset}",
            )
            if assembly_name:
                construction_names.append(assembly_name)

        return sorted(construction_names)

    def get_constructor_r_si_type(self, _row_num: int) -> str:
        """Return "Wall", "Roof" or "Floor" depending on the constructor type."""
        col = self.shape.constructor.inputs.r_si.column
        row_num = _row_num + self.shape.constructor.inputs.rsi_row_offset
        return str(self.xl.get_data(self.shape.name, f"{col}{row_num}"))

    def get_constructor_r_se_type(self, _row_num: int) -> str:
        """Return "Ground", "Outdoor air" or "Ventilated" depending on the constructor type."""
        col = self.shape.constructor.inputs.r_se.column
        row_num = _row_num + self.shape.constructor.inputs.rse_row_offset
        return str(self.xl.get_data(self.shape.name, f"{col}{row_num}"))

    def get_constructor_name(self, _row_num: int) -> str:
        """Return the name of the constructor at the specified row."""
        col = self.shape.constructor.inputs.display_name.column
        row_num = _row_num + self.shape.constructor.inputs.name_row_offset
        return str(self.xl.get_data(self.shape.name, f"{col}{row_num}"))

    def get_first_empty_constructor_start_row(self) -> int:
        """Return the first empty constructor's row number."""
        for row_num in self.all_constructor_start_rows:
            name = self.get_constructor_name(row_num)
            if not self.get_constructor_name(row_num) or name == "None":
                return row_num
        else:
            raise NoEmptyConstructorError

    def get_constructor_u_value(self, _row_num: int) -> Unit:
        """Return the U-Value of the constructor at the specified row."""
        col = self.shape.constructor.inputs.result_val_col
        row_num = _row_num + self.shape.constructor.inputs.result_val_row_offset
        value = float(self.xl.get_data(self.shape.name, f"{col}{row_num}"))
        unit_type = self.shape.constructor.inputs.result_val_unit

        #  -- Handle the fact it might be SI or IP units
        if unit_type == "W/M2K":
            # -- It is already a U-Value
            return Unit(value, unit_type)
        else:
            # -- It is an R-Value, so return the inverse
            return Unit(value, unit_type).inverse()

    def get_constructor_r_value(self, _row_num: int) -> Unit:
        """Return the U-Value of the constructor at the specified row."""
        col = self.shape.constructor.inputs.result_val_col
        row_num = _row_num + self.shape.constructor.inputs.result_val_row_offset
        value = float(self.xl.get_data(self.shape.name, f"{col}{row_num}"))
        unit_type = self.shape.constructor.inputs.result_val_unit

        #  -- Handle the fact it might be SI or IP units
        if unit_type == "W/M2K":
            # -- It is an U-Value, so return the inverse
            return Unit(value, unit_type).inverse()
        else:
            # -- It is already an R-Value
            return Unit(value, unit_type)

    # -------------------------------------------------------------------------
    # -- Writers

    def write_single_PHX_construction(self, _phx_construction, _start_row) -> None:
        """Write a single PHX Construction to the PHPP worksheet."""
        new_constructor = uvalues_constructor.ConstructorBlock(self.shape, _phx_construction)
        self.write_single_constructor_block(new_constructor, _start_row)

    def write_single_constructor_block(
        self, _construction: uvalues_constructor.ConstructorBlock, _start_row: int
    ) -> None:
        """Write a single Construction with all the layers to the PHPP worksheet."""
        for item in _construction.create_xl_items(self.shape.name, _start_row):
            self.xl.write_xl_item(item)

    def write_constructor_blocks(self, _const_blocks: List[uvalues_constructor.ConstructorBlock]) -> None:
        """Write a list of ConstructorBlocks to the U-Values worksheet."""

        self.clear_all_constructor_data(_clear_name=True)

        assert len(_const_blocks) <= len(
            self.all_constructor_start_rows
        ), f"Error: Too many U-Value Constructions: {len(self.all_constructor_start_rows)}"

        for construction, start_row in zip(_const_blocks, self.all_constructor_start_rows):
            self.write_single_constructor_block(construction, start_row)

    def add_new_phx_construction(self, _phx_construction: PhxConstructionOpaque) -> str:
        """Add a new PHX Construction to the PHPP worksheet in the first empty slot found.

        Returns:
        --------
            * (str): The PHPP-id of the new construction.
        """
        new_constructor = uvalues_constructor.ConstructorBlock(self.shape, _phx_construction)
        self.write_single_constructor_block(new_constructor, self.get_first_empty_constructor_start_row())

        return self.get_constructor_phpp_id_by_name(_phx_construction.display_name)

    # -------------------------------------------------------------------------
    # -- Removers

    def clear_single_constructor_data(self, _row_num, _clear_name: bool = False) -> None:
        """Clears the existing data from a single constructor block."""
        # -- main body data
        data_block_start = f"{self.shape.constructor.inputs.sec_1_description.column}{_row_num + self.shape.constructor.inputs.first_layer_row_offset}"
        data_block_end = f"{self.shape.constructor.inputs.thickness.column}{_row_num + self.shape.constructor.inputs.last_layer_row_offset}"
        data_block_range = f"{data_block_start}:{data_block_end}"
        self.xl.write_xl_item(xl_data.XlItem(self.shape.name, data_block_range, None))

        # -- surface films
        self.xl.write_xl_item(
            xl_data.XlItem(
                self.shape.name,
                f"{self.shape.constructor.inputs.r_si.column}{_row_num + 4}:{self.shape.constructor.inputs.r_se.column}{_row_num + 5}",
                None,
            )
        )

        # -- Assembly Name (Sometimes, like with Variants, will want to keep this)
        if _clear_name:
            self.xl.write_xl_item(
                xl_data.XlItem(
                    self.shape.name,
                    f"{self.shape.constructor.inputs.display_name.column}{_row_num + self.shape.constructor.inputs.name_row_offset}",
                    None,
                )
            )

    def clear_all_constructor_data(self, _clear_name: bool = True) -> None:
        """Remove all of the existing input data from all of the constructors in the PHPP."""
        for row_num in self.all_constructor_start_rows:
            self.clear_single_constructor_data(row_num, _clear_name)

    # -------------------------------------------------------------------------

    def activate_variants(self, _assembly_phpp_ids: List[VariantAssemblyLayerName]) -> None:
        """Connect all the links to make the 'Variants' page drive the input values."""

        self.clear_all_constructor_data(_clear_name=False)

        name_col = self.shape.constructor.inputs.display_name.column

        for row_num in self.all_constructor_start_rows:
            assembly_name = self.xl.get_data(
                _sheet_name=self.shape.name,
                _range=f"{name_col}{row_num + self.shape.constructor.inputs.name_row_offset}",
            )

            # -- Find the matching PHPP-ID name from the Variants
            # -- worksheet and make the link
            for variant_layer_id in _assembly_phpp_ids:
                if variant_layer_id.display_name == assembly_name:
                    # -- Link to the variants layer name
                    self.xl.write_xl_item(
                        xl_data.XlItem(
                            self.shape.name,
                            f"{self.shape.constructor.inputs.variants_layer_name}{row_num + self.shape.constructor.inputs.first_layer_row_offset}",
                            variant_layer_id.phpp_id,
                        )
                    )

                    # -- Link the variants layer conductivity
                    self.xl.write_xl_item(
                        xl_data.XlItem(
                            self.shape.name,
                            f"{self.shape.constructor.inputs.sec_1_conductivity.column}{row_num + self.shape.constructor.inputs.first_layer_row_offset}",
                            f"={self.shape.constructor.inputs.variants_conductivity}{row_num + self.shape.constructor.inputs.first_layer_row_offset}",
                        )
                    )

                    # -- Link the variants layer thickness
                    self.xl.write_xl_item(
                        xl_data.XlItem(
                            self.shape.name,
                            f"{self.shape.constructor.inputs.thickness.column}{row_num + self.shape.constructor.inputs.first_layer_row_offset}",
                            f"={self.shape.constructor.inputs.variants_thickness}{row_num + self.shape.constructor.inputs.first_layer_row_offset}",
                        )
                    )

                    # -- Set the surface films to zero
                    self.xl.write_xl_item(
                        xl_data.XlItem(
                            self.shape.name,
                            f"{self.shape.constructor.inputs.r_si.column}{row_num + self.shape.constructor.inputs.rse_row_offset}:{self.shape.constructor.inputs.r_se.column}{row_num + self.shape.constructor.inputs.rsi_row_offset}",
                            0,
                        )
                    )

    def get_all_envelope_assemblies(self) -> List[ExistingAssemblyData]:
        assemblies: List[ExistingAssemblyData] = []

        for row_num in self.used_constructor_start_rows:
            existing_assembly = ExistingAssemblyData(
                self.get_constructor_name(row_num),
                self.get_constructor_u_value(row_num),
                self.get_constructor_r_value(row_num),
                self.get_constructor_r_se_type(row_num),
                self.get_constructor_r_si_type(row_num),
            )
            assemblies.append(existing_assembly)

        return assemblies



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Collection class for Organizing Space Schedules as Utilization Patterns."""

from __future__ import annotations

import uuid
from dataclasses import dataclass, field
from typing import Any, Dict, Generator, ItemsView, KeysView, Optional, Union, ValuesView

from PHX.model.schedules import lighting, occupancy, ventilation


@dataclass
class UtilizationPatternCollection_Ventilation:
    patterns: Dict[Union[str, uuid.UUID], ventilation.PhxScheduleVentilation] = field(init=False, default_factory=dict)

    def __getitem__(self, key) -> ventilation.PhxScheduleVentilation:
        return self.patterns[key]

    def __setitem__(self, key: Union[str, uuid.UUID], value: ventilation.PhxScheduleVentilation) -> None:
        self.patterns[key] = value

    def add_new_util_pattern(self, _util_pattern: Optional[ventilation.PhxScheduleVentilation]) -> None:
        """Add a new ventilation.PhxScheduleVentilation to the Collection.

        Arguments:
        ----------
            * _util_pattern (ventilation.PhxScheduleVentilation): The ventilation.PhxScheduleVentilation
                pattern to add to the collection.

        Returns:
        --------
            * None
        """
        if _util_pattern is None:
            return

        self.patterns[_util_pattern.identifier] = _util_pattern

    def key_is_in_collection(self, _id) -> bool:
        """Check if the id is in the collection."""
        return _id in self.patterns.keys()

    def get_pattern_by_id_num(self, _id_num: int) -> ventilation.PhxScheduleVentilation:
        """Return a ventilation.PhxScheduleVentilation from the collection found by an id-num

        Arguments:
        ----------
            * _id_num (int): The id-number of the Schedule to find.

        Returns:
        --------
            * None
        """
        for pattern in self.patterns.values():
            if pattern.id_num == _id_num:
                return pattern
        msg = f"Error: Cannot locate the ventilation.PhxScheduleVentilation with id-number: {_id_num}"
        raise Exception(msg)

    def __len__(self) -> int:
        return len(self.patterns.keys())

    def __iter__(self) -> Generator[ventilation.PhxScheduleVentilation, Any, None]:
        for v in self.patterns.values():
            yield v

    def __bool__(self) -> bool:
        return bool(self.patterns)

    def items(self) -> ItemsView:
        return self.patterns.items()

    def keys(self) -> KeysView:
        return self.patterns.keys()

    def values(self) -> ValuesView:
        return self.patterns.values()


@dataclass
class UtilizationPatternCollection_Occupancy:
    patterns: Dict[Union[str, uuid.UUID], occupancy.PhxScheduleOccupancy] = field(init=False, default_factory=dict)

    def __getitem__(self, key) -> occupancy.PhxScheduleOccupancy:
        return self.patterns[key]

    def __setitem__(self, key: Union[str, uuid.UUID], value: occupancy.PhxScheduleOccupancy) -> None:
        self.patterns[key] = value

    def add_new_util_pattern(self, _util_pattern: Optional[occupancy.PhxScheduleOccupancy]) -> None:
        """Add a new occupancy.PhxScheduleOccupancy to the Collection.

        Arguments:
        ----------
            * _util_pattern (Optional[occupancy.PhxScheduleOccupancy]): The occupancy.PhxScheduleOccupancy
                pattern to add to the collection.

        Returns:
        --------
            * None
        """
        if _util_pattern is None:
            return

        self.patterns[_util_pattern.identifier] = _util_pattern

    def key_is_in_collection(self, _id: Union[str, uuid.UUID]) -> bool:
        """Check if the id is in the collection."""
        return _id in self.patterns.keys()

    def get_pattern_by_id_num(self, _id_num: int) -> occupancy.PhxScheduleOccupancy:
        """Return a occupancy.PhxScheduleOccupancy from the collection found by an id-num

        Arguments:
        ----------
            * _id_num (int): The id-number of the Schedule to find.

        Returns:
        --------
            * None
        """
        for pattern in self.patterns.values():
            if pattern.id_num == _id_num:
                return pattern
        msg = f"Error: Cannot locate the occupancy.PhxScheduleOccupancy with id-number: {_id_num}"
        raise Exception(msg)

    def __len__(self) -> int:
        return len(self.patterns.keys())

    def __iter__(self) -> Generator[occupancy.PhxScheduleOccupancy, Any, None]:
        for v in self.patterns.values():
            yield v

    def __bool__(self) -> bool:
        return bool(self.patterns)

    def items(self):
        return self.patterns.items()

    def keys(self):
        return self.patterns.keys()

    def values(self):
        return self.patterns.values()


@dataclass
class UtilizationPatternCollection_Lighting:
    patterns: Dict[Union[str, uuid.UUID], lighting.PhxScheduleLighting] = field(init=False, default_factory=dict)

    def __getitem__(self, key) -> lighting.PhxScheduleLighting:
        return self.patterns[key]

    def __setitem__(self, key: Union[str, uuid.UUID], value: lighting.PhxScheduleLighting) -> None:
        self.patterns[key] = value

    def add_new_util_pattern(self, _util_pattern: Optional[lighting.PhxScheduleLighting]) -> None:
        """Add a new lighting.PhxScheduleLighting to the Collection.

        Arguments:
        ----------
            * _util_pattern (Optional[lighting.PhxScheduleLighting]): The lighting.PhxScheduleLighting
                pattern to add to the collection.

        Returns:
        --------
            * None
        """
        if _util_pattern is None:
            return

        self.patterns[_util_pattern.identifier] = _util_pattern

    def key_is_in_collection(self, _id) -> bool:
        """Check if the id is in the collection."""
        return _id in self.patterns.keys()

    def get_pattern_by_id_num(self, _id_num: int) -> lighting.PhxScheduleLighting:
        """Return a lighting.PhxScheduleLighting from the collection found by an id-num

        Arguments:
        ----------
            * _id_num (int): The id-number of the Schedule to find.

        Returns:
        --------
            * None
        """
        for pattern in self.patterns.values():
            if pattern.id_num == _id_num:
                return pattern
        msg = f"Error: Cannot locate the lighting.PhxScheduleLighting with id-number: {_id_num}"
        raise Exception(msg)

    def __len__(self) -> int:
        return len(self.patterns.keys())

    def __iter__(self) -> Generator[lighting.PhxScheduleLighting, Any, None]:
        for v in self.patterns.values():
            yield v

    def __bool__(self) -> bool:
        return bool(self.patterns)

    def items(self):
        return self.patterns.items()

    def keys(self):
        return self.patterns.keys()

    def values(self):
        return self.patterns.values()



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""DEV SANDBOX: convert a WUFI XML file over to HBJSON."""

import pathlib

from rich import print

from PHX.from_WUFI_XML.phx_converter import convert_WUFI_XML_to_PHX_project
from PHX.from_WUFI_XML.read_WUFI_XML_file import get_WUFI_XML_file_as_dict
from PHX.from_WUFI_XML.wufi_file_schema import WUFIplusProject
from PHX.to_WUFI_XML import xml_builder, xml_txt_to_file

SOURCE_DIR = pathlib.Path("tests", "_test_reference_files_xml")
SOURCE_FILE_NAMES = ["School.xml", "_la_mora.xml", "_ridgeway.xml"]
TARGET_DIR = pathlib.Path("tests", "_regenerated_xml")

print(" -" * 50)
for i, xm_source_file_name in enumerate(SOURCE_FILE_NAMES):
    xm_source_file = SOURCE_DIR / xm_source_file_name

    # ----------------------------------------------------------------
    # -- 1) Read in the WUFI-XML File as a new Pydantic Model
    print(f"[green bold]> Reading in data from XML-File: {xm_source_file}[/green bold]")
    wufi_xml_data = get_WUFI_XML_file_as_dict(xm_source_file)
    wufi_xml_model = WUFIplusProject.parse_obj(wufi_xml_data)

    # ----------------------------------------------------------------
    # -- 2) Convert the Pydantic WUFI model over to a PHX model
    print(f"[green bold]> Converting XML-data to a PHX-Model[/green bold]")
    phx_project = convert_WUFI_XML_to_PHX_project(wufi_xml_model)

    # ----------------------------------------------------------------
    # -- 3) Output the PHX model back to a WUFI-XML
    target_file = TARGET_DIR / xm_source_file_name
    xml_txt = xml_builder.generate_WUFI_XML_from_object(phx_project)

    # ----------------------------------------------------------------
    # -- 4) Save the XML file
    print(f"[bold]> Saving the XML file to: ./{target_file}[/bold]")
    xml_txt_to_file.write_XML_text_file(target_file, xml_txt, False)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""XL-App Protocol Classes."""

import pathlib
from typing import Any, Callable, Dict, Optional, Tuple

from PHX.xl import xl_data


class xl_Range_Font:
    def __init__(self):
        self.color: Optional[Tuple[int, ...]]


class xl_RangeColumns_Protocol:
    def __init__(self): ...

    def __len__(self) -> int: ...


class xl_CellRange_Protocol:
    def __init__(self):
        self.address: str


class xl_Range_Protocol:
    def __init__(self):
        self.value: xl_data.xl_range_value
        self.color: Optional[Tuple[int, ...]]
        self.font: xl_Range_Font = xl_Range_Font()
        self.columns: xl_RangeColumns_Protocol = xl_RangeColumns_Protocol()
        self.last_cell: xl_CellRange_Protocol
        self.row: int
        self.column: int
        self.rows: str

    def end(self, *args, **kwargs) -> "xl_Range_Protocol":
        return xl_Range_Protocol()

    def options(self, *args, **kwargs) -> "xl_Range_Protocol":
        return self

    def offset(self, row_offset: int = 0, column_offset: int = 0) -> "xl_Range_Protocol":
        return xl_Range_Protocol()


class xl_API_Protocol:
    def __init__(self, sheet):
        self.sheet: "xl_Sheet_Protocol" = sheet
        self.rows: Dict
        self.outline_object: Any

    def Rows(*args, **kwargs) -> Any: ...

    def unprotect(self) -> None:
        self.sheet.protected = False

    def Unprotect(self) -> None:
        self.sheet.protected = False


class xl_Sheet_Protocol:
    def __init__(self, name="Sheet1"):
        self.api = xl_API_Protocol(self)
        self.protected = True
        self._ranges: Dict[str, xl_Range_Protocol] = {}
        self.name: str = name

    def range(self, cell1: str, cell2: Optional[str] = None) -> xl_Range_Protocol:
        if cell1 in self._ranges.keys():
            return self._ranges[cell1]
        else:
            rng_obj = xl_Range_Protocol()
            self._ranges[cell1] = rng_obj
            return rng_obj

    def clear_contents(self) -> None: ...

    def clear_formats(self) -> None: ...

    def clear(self) -> None: ...

    def activate(self) -> None: ...

    def autofit(self, *args, **kwargs) -> None: ...

    def delete(self) -> None: ...


class xl_Sheets_Protocol:
    def __init__(self):
        self.storage: Dict[str, xl_Sheet_Protocol] = {}

    def __getitem__(self, _key) -> xl_Sheet_Protocol:
        return self.storage[_key]

    def add(
        self,
        name: Optional[str] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
    ) -> xl_Sheet_Protocol:
        if name in self.storage.keys():
            return self.storage[str(name)]
        else:
            new_sheet = xl_Sheet_Protocol(name=str(name))
            self.storage[str(name)] = new_sheet
            return new_sheet

    def __iter__(self):
        for _ in self.storage.values():
            yield _

    def __contains__(self, key):
        return key in self.storage

    def __len__(self):
        return len(self.storage)


class xl_Book_Protocol:
    def __init__(self):
        self.name: str = ""
        self.fullname: str = ""
        self.app = xl_app_Protocol()
        self.sheets = xl_Sheets_Protocol()


class xl_Books_Protocol:
    def __init__(self):
        self.active = xl_Book_Protocol()
        self.count: int = 1

    def add(self) -> xl_Book_Protocol: ...

    def open(self, _p: pathlib.Path) -> xl_Book_Protocol: ...


class xl_app_Protocol:
    def __init__(self):
        self.screen_updating: bool = True
        self.display_alerts: bool = True
        self.calculation: str = "automatic"
        self.visible: bool = False

    def calculate(self) -> None:
        return None


class xl_apps_Protocol:
    def __init__(self):
        self.count: int = 1

    def add(self) -> xl_app_Protocol: ...


class xl_Framework_Protocol:
    def __init__(self):
        self.books: xl_Books_Protocol  # Mac
        self.Books: xl_Books_Protocol  # PC
        self.apps: xl_apps_Protocol  # Mac
        self.Apps: xl_apps_Protocol  # PC

    def Range(self, *args, **kwargs) -> xl_Range_Protocol:
        return xl_Range_Protocol()



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Conversion Schemas for how to write PH/HB objects to WUFI XML"""

import operator
import sys
from functools import reduce
from typing import Any, Dict, List, Optional, TypeVar

from ph_units.converter import convert

from PHX.model import (
    building,
    certification,
    components,
    constructions,
    elec_equip,
    geometry,
    ground,
    hvac,
    phx_site,
    project,
    shades,
    spaces,
)
from PHX.model.enums.foundations import FoundationType
from PHX.model.enums.hvac import PhxHotWaterPipingInchDiameterType
from PHX.model.hvac import _base, renewable_devices
from PHX.model.schedules import occupancy, ventilation
from PHX.to_WUFI_XML.xml_writables import XML_List, XML_Node, XML_Object, xml_writable

TOL_LEV1 = 2  # Value tolerance for rounding. ie; 9.843181919194 -> 9.84
TOL_LEV2 = 10  # Value tolerance for rounding. ie; 9.843181919194 -> 9.8431819192


# -- PROJECT ------------------------------------------------------------------


def _PhxProject(_wufi_project: project.PhxProject) -> List[xml_writable]:
    return [
        XML_Node("DataVersion", _wufi_project.data_version),
        XML_Node("UnitSystem", _wufi_project.unit_system),
        XML_Node("ProgramVersion", _wufi_project.program_version),
        XML_Node("Scope", _wufi_project.scope),
        XML_Node("DimensionsVisualizedGeometry", _wufi_project.visualized_geometry),
        XML_Object("ProjectData", _wufi_project.project_data),
        XML_List(
            "UtilisationPatternsVentilation",
            [
                XML_Object(
                    _node_name="UtilizationPatternVent",
                    _node_object=vent_pat,
                    _attr_name="index",
                    _attr_value=i,
                    _schema_name="_UtilizationPatternVent",
                )
                for i, vent_pat in enumerate(_wufi_project.utilization_patterns_ventilation)
            ],
        ),
        XML_List(
            "UtilizationPatternsPH",
            [
                XML_Object(
                    _node_name="UtilizationPattern",
                    _node_object=util_pat,
                    _attr_name="index",
                    _attr_value=i,
                    _schema_name="_UtilizationPattern",
                )
                for i, util_pat in enumerate(_wufi_project.utilization_patterns_occupancy)
            ],
        ),
        XML_List(
            "Variants",
            [XML_Object("Variant", var, "index", i) for i, var in enumerate(_wufi_project.variants)],
        ),
        XML_List(
            "Assemblies",
            [
                XML_Object("Assembly", at_id, "index", i)
                for i, at_id in enumerate(_wufi_project.assembly_types.values())
            ],
        ),
        XML_List(
            "WindowTypes",
            [
                XML_Object("WindowType", wt_id, "index", i)
                for i, wt_id in enumerate(_wufi_project.window_types.values())
            ],
        ),
        XML_List(
            "SolarProtectionTypes",
            [
                XML_Object(
                    "SolarProtectionType",
                    type_id,
                    "index",
                    i,
                    _schema_name="_PhxWindowShade",
                )
                for i, type_id in enumerate(_wufi_project.shade_types.values())
            ],
        ),
    ]


def _PhxVariant(_variant: project.PhxVariant) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _variant.id_num),
        XML_Node("Name", _variant.name),
        XML_Node("Remarks", _variant.remarks),
        # XML_Node("PlugIn", _variant.plugin),
        XML_Object("Graphics_3D", _variant.graphics3D),
        XML_Object("Building", _variant.building),
        XML_Object("ClimateLocation", _variant.site),
        XML_Object("PassivehouseData", _variant.phius_cert),
        XML_Object("HVAC", _variant._mech_collections, _schema_name="_Systems"),
    ]


def _Systems(
    _collections: List[hvac.PhxMechanicalSystemCollection],
) -> List[xml_writable]:
    return [
        XML_List(
            "Systems",
            [
                XML_Object(
                    "System",
                    collection,
                    "index",
                    i,
                    _schema_name="_PhxMechanicalSystemCollection",
                )
                for i, collection in enumerate(_collections)
            ],
        )
    ]


def _PhxDateProject(_project_data: project.PhxProjectDate) -> List[xml_writable]:
    return [
        XML_Node("Year", _project_data.year),
        XML_Node("Month", _project_data.month),
        XML_Node("Day", _project_data.day),
        XML_Node("Hour", _project_data.hour),
        XML_Node("Minutes", _project_data.minutes),
    ]


def _PhxProjectData(_project_data: project.PhxProjectData) -> List[xml_writable]:
    return [
        XML_Node("Year_Construction", _project_data.year_constructed),
        XML_Node("OwnerIsClient", _project_data.owner_is_client),
        XML_Object("Date_Project", _project_data.project_date, _schema_name="_PhxDateProject"),
        XML_Node("WhiteBackgroundPictureBuilding", _project_data.image),
        XML_Node("Customer_Name", _project_data.customer.name),
        XML_Node("Customer_Street", _project_data.customer.street),
        XML_Node("Customer_Locality", _project_data.customer.city),
        XML_Node("Customer_PostalCode", _project_data.customer.post_code),
        XML_Node("Customer_Tel", _project_data.customer.telephone),
        XML_Node("Customer_Email", _project_data.customer.email),
        XML_Node("Building_Name", _project_data.building.name),
        XML_Node("Building_Street", _project_data.building.street),
        XML_Node("Building_Locality", _project_data.building.city),
        XML_Node("Building_PostalCode", _project_data.building.post_code),
        XML_Node("Owner_Name", _project_data.owner.name),
        XML_Node("Owner_Street", _project_data.owner.street),
        XML_Node("Owner_Locality", _project_data.owner.city),
        XML_Node("Owner_PostalCode", _project_data.owner.post_code),
        XML_Node("Responsible_Name", _project_data.designer.name),
        XML_Node("Responsible_Street", _project_data.designer.street),
        XML_Node("Responsible_Locality", _project_data.designer.city),
        XML_Node("Responsible_PostalCode", _project_data.designer.post_code),
        XML_Node("Responsible_Tel", _project_data.designer.telephone),
        XML_Node("Responsible_LicenseNr", _project_data.designer.license_number),
        XML_Node("Responsible_Email", _project_data.designer.email),
    ]


# -- BUILDING -----------------------------------------------------------------


def _PhxBuilding(_b: building.PhxBuilding) -> List[xml_writable]:
    return [
        XML_List(
            "Components",
            [XML_Object("Component", c, "index", i) for i, c in enumerate(_b.all_components)],
        ),
        XML_List("Zones", [XML_Object("Zone", z, "index", i) for i, z in enumerate(_b.zones)]),
    ]


def _PhxZone(_z: building.PhxZone) -> List[xml_writable]:
    _spec_cap_WH_m2k = {
        1: 60,
        2: 132,
        3: 204,
    }

    def wufi_spaces(_z: building.PhxZone) -> List[spaces.PhxSpace]:
        """Return a list of all the spaces in the PhxZone for reporting out to WUFI."""
        if _z.merge_spaces_by_erv == False:
            # -- Return all the spaces in the zone with ventilation airflow
            return _z.ventilated_spaces
        else:
            # -- Merge the Spaces together by their ERV ID
            merged_spaces_: list[spaces.PhxSpace] = []
            for space_group in _z.ventilated_spaces_grouped_by_erv:
                if len(space_group) > 1:
                    new_space = reduce(operator.add, space_group)
                else:
                    # -- If there is only 1 space, won't go through the __add__
                    # -- so be sure to set up the display_name manually here
                    new_space = space_group[0]
                    new_space.display_name = new_space.vent_unit_display_name
                merged_spaces_.append(new_space)
            return sorted(merged_spaces_, key=lambda x: x.vent_unit_display_name)

    return [
        XML_Node("Name", _z.display_name),
        XML_Node("KindZone", _z.zone_type.value),
        XML_Node("KindAttachedZone", _z.attached_zone_type.value),
        XML_Node("TemperatureReductionFactorUserDefined", _z.attached_zone_reduction_factor),
        XML_Node("IdentNr", _z.id_num),
        XML_List(
            "RoomsVentilation",
            [XML_Object("Room", sp, "index", i, _schema_name="_PhxSpace") for i, sp in enumerate(wufi_spaces(_z))],
        ),
        XML_List(
            "LoadsPersonsPH",
            [XML_Object("LoadPerson", sp, "index", i, _schema_name="_LoadPerson") for i, sp in enumerate(_z.spaces)],
        ),
        XML_List(
            "LoadsLightingsPH",
            [
                XML_Object("LoadsLighting", sp, "index", i, _schema_name="_LoadsLighting")
                for i, sp in enumerate(_z.spaces)
            ],
        ),
        XML_Node("GrossVolume_Selection", _z.volume_gross_mode.value),
        XML_Node("GrossVolume", _z.volume_gross),
        XML_Node("NetVolume_Selection", _z.volume_net_mode.value),
        XML_Node("NetVolume", _z.volume_net),
        XML_Node("FloorArea_Selection", _z.weighted_floor_area_mode.value),
        XML_Node("FloorArea", _z.icfa_override or _z.weighted_net_floor_area),
        XML_Node("ClearanceHeight_Selection", 1),
        XML_Node("ClearanceHeight", _z.clearance_height),
        XML_Node("SpecificHeatCapacity_Selection", _z.specific_heat_capacity.value),
        XML_Node("SpecificHeatCapacity", _spec_cap_WH_m2k[_z.specific_heat_capacity.value]),
        XML_Node("IdentNrPH_Building", 1),
        XML_Node("OccupantQuantityUserDef", int(_z.res_occupant_quantity), "unit", "-"),
        XML_Node("NumberBedrooms", int(_z.res_number_bedrooms), "unit", "-"),
        XML_List(
            "HomeDevice",
            [
                XML_Object("Device", d, "index", i, _schema_name="_PhxElectricalDevice")
                for i, d in enumerate(_z.elec_equipment_collection.devices)
            ],
        ),
        XML_List(
            "ExhaustVents",
            [
                XML_Object("ExhaustVent", v, "index", i, _schema_name="_PhxExhaustVentilator")
                for i, v in enumerate(_z.exhaust_ventilator_collection.devices)
            ],
        ),
        XML_List(
            "ThermalBridges",
            [
                XML_Object("ThermalBridge", tb, "index", i, _schema_name="_PhxThermalBridge")
                for i, tb in enumerate(_z.thermal_bridges)
            ],
        ),
        XML_Node("SummerNaturalVentilationDay", 0, "unit", "1/hr"),
        XML_Node("SummerNaturalVentilationNight", 0, "unit", "1/hr"),
    ]


def _PhxThermalBridge(_tb: components.PhxComponentThermalBridge) -> List[xml_writable]:
    return [
        XML_Node("Name", _tb.display_name),
        XML_Node("Type", _tb.group_number * -1),
        XML_Node("Length", _tb.length),
        XML_Node("PsiValue", _tb.psi_value),
        XML_Node("IdentNrOptionalClimate", -1),
    ]


def _PhxComponentOpaque(_c: components.PhxComponentOpaque) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _c.id_num),
        XML_Node("Name", _c.display_name),
        XML_Node("Visual", True),
        XML_Node("Type", _c.face_opacity.value),
        XML_Node("IdentNrColorI", _c.color_interior.value),
        XML_Node("IdentNrColorE", _c.color_exterior.value),
        XML_Node("InnerAttachment", _c.exposure_interior),
        XML_Node("OuterAttachment", _c.exposure_exterior.value),
        XML_Node("IdentNr_ComponentInnerSurface", _c.interior_attachment_id),
        XML_Node("IdentNrAssembly", _c.assembly_type_id_num),
        XML_Node("IdentNrWindowType", -1),
        XML_List(
            "IdentNrPolygons",
            [XML_Node("IdentNr", n, "index", i) for i, n in enumerate(_c.polygon_ids)],
        ),
    ]


def _PhxComponentAperture(_c: components.PhxComponentAperture) -> List[xml_writable]:
    # Use the shading, if set, otherwise use the Top-Frame width
    shading_reveal_distance = _c.average_shading_d_reveal or _c.window_type.frame_top.width

    return [
        XML_Node("IdentNr", _c.id_num),
        XML_Node("Name", _c.display_name),
        XML_Node("Visual", True),
        XML_Node("Type", _c.face_opacity.value),  # <--- WUFI uses this instead of FaceType
        XML_Node("IdentNrColorI", _c.color_interior.value),
        XML_Node("IdentNrColorE", _c.color_exterior.value),
        XML_Node("InnerAttachment", _c.exposure_interior),
        XML_Node("OuterAttachment", _c.exposure_exterior.value),
        XML_Node("IdentNr_ComponentInnerSurface", _c.interior_attachment_id),
        XML_Node("IdentNrAssembly", -1),
        XML_Node("IdentNrWindowType", _c.window_type_id_num),
        XML_List(
            "IdentNrPolygons",
            [XML_Node("IdentNr", n, "index", i) for i, n in enumerate(_c.polygon_ids)],
        ),
        XML_Node("DepthWindowReveal", _c._install_depth, "unit", "m"),
        XML_Node("DistanceDaylightOpeningToReveal", shading_reveal_distance, "unit", "m"),
        XML_Node("IdentNrSolarProtection", _c.shade_type_id_num),
        XML_Node("IdentNrOverhang", -1),
        XML_Node("DefaultCorrectionShadingMonth", _c.default_monthly_shading_correction_factor),
    ]


# -- CERTIFICATION ------------------------------------------------------------


def _PhxPhBuildingData(
    _phius_cert: certification.PhxPhiusCertification,
) -> List[xml_writable]:
    # alias --
    bd = _phius_cert.ph_building_data
    cert = _phius_cert.phius_certification_settings

    return [
        XML_Node("IdentNr", bd._count),
        XML_Node("BuildingCategory", cert.phius_building_category_type.value),
        XML_Node("OccupancyTypeResidential", cert.phius_building_use_type.value),
        XML_Node("BuildingStatus", cert.phius_building_status.value),
        XML_Node("BuildingType", cert.phius_building_type.value),
        XML_Node("OccupancySettingMethod", bd.occupancy_setting_method),
        XML_Node("NumberUnits", bd.num_of_units),
        XML_Node("CountStories", bd.num_of_floors),
        XML_Node("EnvelopeAirtightnessCoefficient", bd.airtightness_q50),
        XML_Node("SummerHRVHumidityRecovery", bd.summer_hrv_bypass_mode.value),
        XML_Node("BuildingWindExposure", bd.building_exposure_type.value),
        XML_List(
            "FoundationInterfaces",
            [
                XML_Object("FoundationInterface", f, "index", i, _schema_name="_PhxFoundation")
                for i, f in enumerate(bd.foundations)
            ],
        ),
        XML_Object(
            "InternalGainsAdditionalData",
            None,
            _schema_name="_InternalGainsAdditionalData",
        ),
        XML_Node("MechanicalRoomTemperature", bd.mech_room_temp),
        XML_Node("IndoorTemperature", bd.setpoints.winter),
        XML_Node("OverheatingTemperatureThreshold", bd.setpoints.summer),
        XML_Node("NonCombustibleMaterials", bd.non_combustible_materials),
    ]


def _InternalGainsAdditionalData(*args, **kwargs) -> List[xml_writable]:
    return [
        XML_Node("EvaporationHeatPerPerson", 15, "unit", "W"),
        XML_Node("HeatLossFluschingWC", True),
        XML_Node("QuantityWCs", 1),
        XML_Node("RoomCategory", 1),
        XML_Node("UseDefaultValuesSchool", False),
        XML_Node("MarginalPerformanceRatioDHW", None),
    ]


def _PhxPhiusCertification(
    _phius_cert: certification.PhxPhiusCertification,
) -> List[xml_writable]:
    # No idea why this is a list in Wufi? When would there ever be more than 1? whatever...
    _temp_bldg_data_list = [_phius_cert]

    return [
        XML_Node(
            "PH_CertificateCriteria",
            _phius_cert.phius_certification_settings.phius_building_certification_program.value,
        ),
        XML_Node(
            "PH_SelectionTargetData",
            _phius_cert.phius_certification_criteria.ph_selection_target_data,
        ),
        XML_Node(
            "AnnualHeatingDemand",
            _phius_cert.phius_certification_criteria.phius_annual_heating_demand,
        ),
        XML_Node(
            "AnnualCoolingDemand",
            _phius_cert.phius_certification_criteria.phius_annual_cooling_demand,
        ),
        XML_Node(
            "PeakHeatingLoad",
            _phius_cert.phius_certification_criteria.phius_peak_heating_load,
        ),
        XML_Node(
            "PeakCoolingLoad",
            _phius_cert.phius_certification_criteria.phius_peak_cooling_load,
        ),
        XML_List(
            "PH_Buildings",
            [
                XML_Object("PH_Building", obj, "index", i, _schema_name="_PhxPhBuildingData")
                for i, obj in enumerate(_temp_bldg_data_list)
            ],
        ),
        XML_Node("UseWUFIMeanMonthShading", _phius_cert.use_monthly_shading),
    ]


# -- FOUNDATIONS --------------------------------------------------------------


def _PhxHeatedBasement(_f: ground.PhxHeatedBasement):
    return [
        XML_Node("FloorSlabArea_Selection", 6),  # 6=User defined
        XML_Node("FloorSlabArea", _f.floor_slab_area_m2),
        XML_Node("FloorSlabPerimeter_Selection", 6),  # 6=User defined
        XML_Node("FloorSlabPerimeter", _f.floor_slab_exposed_perimeter_m),
        XML_Node("U_ValueBasementSlab_Selection", 6),  # 6=User defined
        XML_Node("U_ValueBasementSlab", _f.floor_slab_u_value),
        XML_Node("DepthBasementBelowGroundSurface_Selection", 6),  # 6=user-defined
        XML_Node("DepthBasementBelowGroundSurface", _f.slab_depth_below_grade_m),
        # XML_Node("HeightBasementWallAboveGrade_Selection", 6), #6=user-defined
        # XML_Node("HeightBasementWallAboveGrade", _f.ab),
        XML_Node("U_ValueBasementWall_Selection", 6),  # 6=user-defined
        XML_Node("U_ValueBasementWall", _f.basement_wall_u_value),
    ]


def _PhxUnHeatedBasement(_f: ground.PhxUnHeatedBasement):
    return [
        XML_Node("DepthBasementBelowGroundSurface_Selection", 6),  # 6=User defined
        XML_Node("DepthBasementBelowGroundSurface", _f.slab_depth_below_grade_m),
        XML_Node("HeightBasementWallAboveGrade_Selection", 6),  # 6=User defined
        XML_Node("HeightBasementWallAboveGrade", _f.basement_wall_height_above_grade_m),
        XML_Node("FloorSlabArea_Selection", 6),  # 6=User defined
        XML_Node("FloorSlabArea", _f.floor_ceiling_area_m2),
        XML_Node("U_ValueBasementSlab_Selection", 6),  # 6=User defined
        XML_Node("U_ValueBasementSlab", _f.floor_slab_u_value),
        XML_Node("FloorCeilingArea_Selection", 6),  # 6=User defined
        XML_Node("FloorCeilingArea", _f.floor_ceiling_area_m2),
        XML_Node("U_ValueCeilingToUnheatedCellar_Selection", 6),  # 6=User defined
        XML_Node("U_ValueCeilingToUnheatedCellar", _f.ceiling_u_value),
        XML_Node("U_ValueBasementWall_Selection", 6),  # 6=User defined
        XML_Node("U_ValueBasementWall", _f.basement_wall_uValue_below_grade),
        XML_Node("U_ValueWallAboveGround_Selection", 6),  # 6=User defined
        XML_Node("U_ValueWallAboveGround", _f.basement_wall_uValue_above_grade),
        XML_Node("FloorSlabPerimeter_Selection", 6),  # 6=User defined
        XML_Node("FloorSlabPerimeter", _f.floor_slab_exposed_perimeter_m),
        XML_Node("BasementVolume_Selection", 6),  # 6=User defined
        XML_Node("BasementVolume", _f.basement_volume_m3),
    ]


def _PhxSlabOnGrade(_f: ground.PhxSlabOnGrade):
    return [
        XML_Node("FloorSlabArea_Selection", 6),  # 6=User defined
        XML_Node("FloorSlabArea", _f.floor_slab_area_m2),
        XML_Node("U_ValueBasementSlab_Selection", 6),  # 6=User defined
        XML_Node("U_ValueBasementSlab", _f.floor_slab_u_value),
        XML_Node("FloorSlabPerimeter_Selection", 6),  # 6=User defined
        XML_Node("FloorSlabPerimeter", _f.floor_slab_exposed_perimeter_m),
        XML_Node("PositionPerimeterInsulation", _f._perim_insulation_position.value),
        XML_Node("PerimeterInsulationWidthDepth", _f.perim_insulation_width_or_depth_m),
        XML_Node("ConductivityPerimeterInsulation", _f.perim_insulation_conductivity),
        XML_Node("ThicknessPerimeterInsulation", _f.perim_insulation_thickness_m),
    ]


def _PhxVentedCrawlspace(_f: ground.PhxVentedCrawlspace):
    return [
        XML_Node("FloorCeilingArea_Selection", 6),  # 6=user-defined
        XML_Node("FloorCeilingArea", _f.crawlspace_floor_slab_area_m2),
        XML_Node("U_ValueCeilingToUnheatedCellar_Selection", 6),  # 6=User defined
        XML_Node("U_ValueCeilingToUnheatedCellar", _f.ceiling_above_crawlspace_u_value),
        XML_Node("FloorSlabPerimeter_Selection", 6),  # 6=User defined
        XML_Node("FloorSlabPerimeter", _f.crawlspace_floor_exposed_perimeter_m),
        XML_Node("HeightBasementWallAboveGrade_Selection", 6),  # 6=User defined
        XML_Node("HeightBasementWallAboveGrade", _f.crawlspace_wall_height_above_grade_m),
        XML_Node("U_ValueCrawlspaceFloor_Selection", 6),  # 6=user-defined
        XML_Node("U_ValueCrawlspaceFloor", _f.crawlspace_floor_u_value),
        XML_Node("CrawlspaceVentOpenings_Selection", 6),  # 6=user-defined
        XML_Node("CrawlspaceVentOpenings", _f.crawlspace_vent_opening_are_m2),
        XML_Node("U_ValueWallAboveGround_Selection", 6),  # 6=user-defined
        XML_Node("U_ValueWallAboveGround", _f.crawlspace_wall_u_value),
    ]


def _PhxFoundation(_f: ground.PhxFoundation) -> List[xml_writable]:
    common_attributes = [
        XML_Node("Name", _f.display_name),
        XML_Node("SettingFloorSlabType", _f.foundation_setting_num.value),
        XML_Node("FloorSlabType", _f.foundation_type_num.value),
    ]

    if _f.foundation_type_num == FoundationType.NONE:
        foundation_specific_attributes: List[xml_writable] = []
    else:
        foundation_schema = getattr(sys.modules[__name__], f"_{_f.__class__.__name__}")
        foundation_specific_attributes = foundation_schema(_f)

    return common_attributes + foundation_specific_attributes


# -- CLIMATE ------------------------------------------------------------------


def _PH_ClimateLocation(_phx_site: phx_site.PhxSite) -> List[xml_writable]:
    def _in_wufi_order(_factor_dict: dict) -> List[phx_site.PhxEnergyFactorAlias]:
        """Returns the PE /CO2 conversion factors in WUFI-specific order."""
        fuel_order = [
            "OIL",
            "NATURAL_GAS",
            "LPG",
            "HARD_COAL",
            "WOOD",
            "ELECTRICITY_MIX",
            "ELECTRICITY_PV",
            "HARD_COAL_CGS_70_CHP",
            "HARD_COAL_CGS_35_CHP",
            "HARD_COAL_CGS_0_CHP",
            "GAS_CGS_70_CHP",
            "GAS_CGS_35_CHP",
            "GAS_CGS_0_CHP",
            "OIL_CGS_70_CHP",
            "OIL_CGS_35_CHP",
            "OIL_CGS_0_CHP",
        ]
        return [_factor_dict[fuel_name] for fuel_name in fuel_order]

    return [
        XML_Node("Selection", _phx_site.climate.selection.value),
        XML_Node("DailyTemperatureSwingSummer", _phx_site.climate.daily_temp_swing),
        XML_Node("AverageWindSpeed", _phx_site.climate.avg_wind_speed),
        # -- Location
        XML_Node("Latitude", _phx_site.location.latitude),
        XML_Node("Longitude", _phx_site.location.longitude),
        XML_Node("HeightNNWeatherStation", _phx_site.climate.station_elevation),
        XML_Node("dUTC", _phx_site.location.hours_from_UTC),
        XML_Node("HeightNNBuilding", _phx_site.location.site_elevation),
        XML_Node("ClimateZone", _phx_site.location.climate_zone),
        # -- Ground
        XML_Node("GroundThermalConductivity", _phx_site.ground.ground_thermal_conductivity),
        XML_Node("GroundHeatCapacitiy", _phx_site.ground.ground_heat_capacity),
        XML_Node("GroundDensity", _phx_site.ground.ground_density),
        XML_Node("DepthGroundwater", _phx_site.ground.depth_groundwater),
        XML_Node("FlowRateGroundwater", _phx_site.ground.flow_rate_groundwater),
        # -- Monthly
        XML_List(
            "TemperatureMonthly",
            [XML_Node("Item", val, "index", i) for i, val in enumerate(_phx_site.climate.temperature_air)],
        ),
        XML_List(
            "DewPointTemperatureMonthly",
            [XML_Node("Item", val, "index", i) for i, val in enumerate(_phx_site.climate.temperature_dewpoint)],
        ),
        XML_List(
            "SkyTemperatureMonthly",
            [XML_Node("Item", val, "index", i) for i, val in enumerate(_phx_site.climate.temperature_sky)],
        ),
        XML_List(
            "NorthSolarRadiationMonthly",
            [XML_Node("Item", val, "index", i) for i, val in enumerate(_phx_site.climate.radiation_north)],
        ),
        XML_List(
            "EastSolarRadiationMonthly",
            [XML_Node("Item", val, "index", i) for i, val in enumerate(_phx_site.climate.radiation_east)],
        ),
        XML_List(
            "SouthSolarRadiationMonthly",
            [XML_Node("Item", val, "index", i) for i, val in enumerate(_phx_site.climate.radiation_south)],
        ),
        XML_List(
            "WestSolarRadiationMonthly",
            [XML_Node("Item", val, "index", i) for i, val in enumerate(_phx_site.climate.radiation_west)],
        ),
        XML_List(
            "GlobalSolarRadiationMonthly",
            [XML_Node("Item", val, "index", i) for i, val in enumerate(_phx_site.climate.radiation_global)],
        ),
        # -- Peak Load Values
        XML_Node("TemperatureHeating1", _phx_site.climate.peak_heating_1.temperature_air),
        XML_Node(
            "NorthSolarRadiationHeating1",
            _phx_site.climate.peak_heating_1.radiation_north,
        ),
        XML_Node("EastSolarRadiationHeating1", _phx_site.climate.peak_heating_1.radiation_east),
        XML_Node(
            "SouthSolarRadiationHeating1",
            _phx_site.climate.peak_heating_1.radiation_south,
        ),
        XML_Node("WestSolarRadiationHeating1", _phx_site.climate.peak_heating_1.radiation_west),
        XML_Node(
            "GlobalSolarRadiationHeating1",
            _phx_site.climate.peak_heating_1.radiation_global,
        ),
        XML_Node("TemperatureHeating2", _phx_site.climate.peak_heating_2.temperature_air),
        XML_Node(
            "NorthSolarRadiationHeating2",
            _phx_site.climate.peak_heating_2.radiation_north,
        ),
        XML_Node("EastSolarRadiationHeating2", _phx_site.climate.peak_heating_2.radiation_east),
        XML_Node(
            "SouthSolarRadiationHeating2",
            _phx_site.climate.peak_heating_2.radiation_south,
        ),
        XML_Node("WestSolarRadiationHeating2", _phx_site.climate.peak_heating_2.radiation_west),
        XML_Node(
            "GlobalSolarRadiationHeating2",
            _phx_site.climate.peak_heating_2.radiation_global,
        ),
        XML_Node("TemperatureCooling", _phx_site.climate.peak_cooling_1.temperature_air),
        XML_Node("NorthSolarRadiationCooling", _phx_site.climate.peak_cooling_1.radiation_north),
        XML_Node("EastSolarRadiationCooling", _phx_site.climate.peak_cooling_1.radiation_east),
        XML_Node("SouthSolarRadiationCooling", _phx_site.climate.peak_cooling_1.radiation_south),
        XML_Node("WestSolarRadiationCooling", _phx_site.climate.peak_cooling_1.radiation_west),
        XML_Node(
            "GlobalSolarRadiationCooling",
            _phx_site.climate.peak_cooling_1.radiation_global,
        ),
        XML_Node("TemperatureCooling2", _phx_site.climate.peak_cooling_2.temperature_air),
        XML_Node(
            "NorthSolarRadiationCooling2",
            _phx_site.climate.peak_cooling_2.radiation_north,
        ),
        XML_Node("EastSolarRadiationCooling2", _phx_site.climate.peak_cooling_2.radiation_east),
        XML_Node(
            "SouthSolarRadiationCooling2",
            _phx_site.climate.peak_cooling_2.radiation_south,
        ),
        XML_Node("WestSolarRadiationCooling2", _phx_site.climate.peak_cooling_2.radiation_west),
        XML_Node(
            "GlobalSolarRadiationCooling2",
            _phx_site.climate.peak_cooling_2.radiation_global,
        ),
        XML_Node("SelectionPECO2Factor", _phx_site.energy_factors.selection_pe_co2_factor.value),
        XML_List(
            "PEFactorsUserDef",
            [
                XML_Node(f"PEF{i}", factor.value, "unit", factor.unit)
                for i, factor in enumerate(_in_wufi_order(_phx_site.energy_factors.pe_factors))
            ],
        ),
        XML_List(
            "CO2FactorsUserDef",
            [
                XML_Node(f"CO2F{i}", factor.value, "unit", factor.unit)
                for i, factor in enumerate(_in_wufi_order(_phx_site.energy_factors.co2_factors))
            ],
        ),
    ]


def _PhxSite(_phx_site: phx_site.PhxSite) -> List[xml_writable]:
    return [
        XML_Node("Selection", _phx_site.selection.value),
        # XML_Node('IDNr_DB', _climate.),
        # XML_Node('Name_DB', _climate.),
        # XML_Node('Comment_DB', _climate.),
        XML_Node("Latitude_DB", _phx_site.location.latitude, "unit", "°"),
        XML_Node("Longitude_DB", _phx_site.location.longitude, "unit", "°"),
        XML_Node("HeightNN_DB", _phx_site.location.site_elevation, "unit", "m"),
        XML_Node("dUTC_DB", _phx_site.location.hours_from_UTC),
        # XML_Node('FileName_DB', _climate.),
        # XML_Node('Type_DB', _climate.),
        # XML_Node('CatalogueNr_DB', _climate.),
        # XML_Node('MapNr_DB', _climate.),
        XML_Node("Albedo", -2, "choice", "User defined"),
        XML_Node("GroundReflShort", 0.2, "unit", "-"),
        XML_Node("GroundReflLong", 0.1, "unit", "-"),
        XML_Node("GroundEmission", 0.9, "unit", "-"),
        XML_Node("CloudIndex", 0.66, "unit", "-"),
        XML_Node("CO2concenration", 350, "unit", "mg/m³"),
        XML_Node("Unit_CO2concentration", 48, "choice", "ppmv"),
        XML_Object("PH_ClimateLocation", _phx_site, _schema_name="_PH_ClimateLocation"),
    ]


# -- GEOMETRY -----------------------------------------------------------------


def _PhxGraphics3D(_graphics3D: geometry.PhxGraphics3D) -> List[xml_writable]:
    return [
        XML_List(
            "Vertices",
            [XML_Object("Vertix", var, "index", i) for i, var in enumerate(_graphics3D.vertices)],
        ),
        XML_List(
            "Polygons",
            [
                XML_Object("Polygon", var, "index", i, _schema_name="_PhxPolygon")
                for i, var in enumerate(_graphics3D.polygons)
            ],
        ),
    ]


def _PhxPolygon(_p: geometry.PhxPolygon) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _p.id_num),
        XML_Node("NormalVectorX", round(_p.normal_vector.x, TOL_LEV2)),
        XML_Node("NormalVectorY", round(_p.normal_vector.y, TOL_LEV2)),
        XML_Node("NormalVectorZ", round(_p.normal_vector.z, TOL_LEV2)),
        XML_List(
            "IdentNrPoints",
            [XML_Node("IdentNr", v_id, "index", i) for i, v_id in enumerate(_p.vertices_id_numbers)],
        ),
        XML_List(
            "IdentNrPolygonsInside",
            [XML_Node("IdentNr", v_id, "index", i) for i, v_id in enumerate(_p.child_polygon_ids)],
        ),
    ]


def _PhxVertix(_v: geometry.PhxVertix) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _v.id_num),
        XML_Node("X", round(_v.x, TOL_LEV2)),
        XML_Node("Y", round(_v.y, TOL_LEV2)),
        XML_Node("Z", round(_v.z, TOL_LEV2)),
    ]


# -- CONSTRUCTIONS ------------------------------------------------------------


def _PhxConstructionOpaque(_a: constructions.PhxConstructionOpaque) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _a.id_num),
        XML_Node("Name", _a.display_name),
        XML_Node("Order_Layers", _a.layer_order),
        XML_Node("Grid_Kind", _a.grid_kind),
        XML_List(
            "Layers",
            [XML_Object("Layer", n, "index", i) for i, n in enumerate(_a.layers)],
        ),
        XML_List(
            "ExchangeMaterials",
            [
                XML_Object("ExchangeMaterial", n, "index", i, _schema_name="_PhxExchangeMaterial")
                for i, n in enumerate(_a.exchange_materials)
            ],
        ),
    ]


def _PhxLayer(_l: constructions.PhxLayer) -> List[xml_writable]:
    # If there is only one row, ignore it
    if len(_l.divisions.row_heights) == 1:
        row_heights = []
    else:
        row_heights = _l.divisions.row_heights

    # if there is only one column, ignore it
    if len(_l.divisions.column_widths) == 1:
        column_widths = []
    else:
        column_widths = _l.divisions.column_widths

    return [
        XML_Node("Thickness", _l.thickness_m),
        XML_Object("Material", _l.material),
        XML_List(
            "ExchangeDivisionHorizontal",
            [
                XML_Object("DivisionH", n, "index", i, _schema_name="_PhxLayerDivision")
                for i, n in enumerate(column_widths)
            ],
        ),
        XML_List(
            "ExchangeDivisionVertical",
            [
                XML_Object("DivisionV", n, "index", i, _schema_name="_PhxLayerDivision")
                for i, n in enumerate(row_heights)
            ],
        ),
        XML_List(
            "ExchangeMaterialIdentNrs",
            [
                XML_Object("MaterialIDNr", m, "index", i, _schema_name="_MaterialIDNr")
                for i, m in enumerate(_l.division_material_id_numbers)
            ],
        ),
    ]


def _PhxLayerDivision(_d: float) -> List[xml_writable]:
    return [
        XML_Node("Distance", _d),
        XML_Node("ExpandingContracting", 2),
    ]


def _MaterialIDNr(_id_num: int) -> List[xml_writable]:
    return [
        XML_Node("Type", 1),
        XML_Node("IdentNr_Object", _id_num),
    ]


def _PhxMaterial(_m: constructions.PhxMaterial) -> List[xml_writable]:
    return [
        XML_Node("Name", _m.display_name),
        XML_Node("ThermalConductivity", _m.conductivity),
        XML_Node("BulkDensity", _m.density),
        XML_Node("Porosity", _m.porosity),
        XML_Node("HeatCapacity", _m.heat_capacity),
        XML_Node("WaterVaporResistance", _m.water_vapor_resistance),
        XML_Node("ReferenceWaterContent", _m.reference_water),
        XML_Object("Color", _m.argb_color, _schema_name="_PhxColor"),
    ]


def _PhxColor(_c: constructions.PhxColor) -> List[xml_writable]:
    return [
        XML_Node("Alpha", _c.alpha),
        XML_Node("Red", _c.red),
        XML_Node("Green", _c.green),
        XML_Node("Blue", _c.blue),
    ]


def _PhxExchangeMaterial(_m: constructions.PhxMaterial) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _m.id_num),
        XML_Node("Name", _m.display_name),
        XML_Node("ThermalConductivity", _m.conductivity),
        XML_Node("BulkDensity", _m.density),
        XML_Node("HeatCapacity", _m.heat_capacity),
        XML_Object("Color", _m.argb_color, _schema_name="_PhxColor"),
    ]


def _PhxConstructionWindow(
    _wt: constructions.PhxConstructionWindow,
) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _wt.id_num),
        XML_Node("Name", _wt.display_name),
        XML_Node("Uw_Detailed", _wt.use_detailed_uw),
        XML_Node("GlazingFrameDetailed", _wt.use_detailed_frame),
        XML_Node("FrameFactor", _wt.frame_factor),
        XML_Node("U_Value", _wt.u_value_window),
        XML_Node("U_Value_Glazing", _wt.u_value_glass),
        XML_Node("MeanEmissivity", _wt.glass_mean_emissivity),
        XML_Node("g_Value", _wt.glass_g_value),
        XML_Node("SHGC_Hemispherical", _wt.glass_g_value),
        # --
        XML_Node("U_Value_Frame", _wt.u_value_frame),
        # --
        XML_Node("Frame_Width_Left", _wt.frame_left.width),
        XML_Node("Frame_Psi_Left", _wt.frame_left.psi_install),
        XML_Node("Frame_U_Left", _wt.frame_left.u_value),
        XML_Node("Glazing_Psi_Left", _wt.frame_left.psi_glazing),
        # --
        XML_Node("Frame_Width_Right", _wt.frame_right.width),
        XML_Node("Frame_Psi_Right", _wt.frame_right.psi_install),
        XML_Node("Frame_U_Right", _wt.frame_right.u_value),
        XML_Node("Glazing_Psi_Right", _wt.frame_right.psi_glazing),
        # --
        XML_Node("Frame_Width_Top", _wt.frame_top.width),
        XML_Node("Frame_Psi_Top", _wt.frame_top.psi_install),
        XML_Node("Frame_U_Top", _wt.frame_top.u_value),
        XML_Node("Glazing_Psi_Top", _wt.frame_top.psi_glazing),
        # --
        XML_Node("Frame_Width_Bottom", _wt.frame_bottom.width),
        XML_Node("Frame_Psi_Bottom", _wt.frame_bottom.psi_install),
        XML_Node("Frame_U_Bottom", _wt.frame_bottom.u_value),
        XML_Node("Glazing_Psi_Bottom", _wt.frame_bottom.psi_glazing),
    ]


def _PhxWindowShade(_s: shades.PhxWindowShade) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _s.id_num),
        XML_Node("Name", _s.display_name),
        XML_Node("OperationMode", _s.operation_mode),
        XML_Node("MaxRedFactorRadiation", _s.reduction_factor),
        XML_Node("ExternalEmissivity", _s.external_emissivity),
        XML_Node("EquivalentAbsorptivity", _s.absorptivity),
        XML_Node("ThermalResistanceSupplement", _s.thermal_resistance_supplement),
        XML_Node("ThermalResistanceCavity", _s.thermal_resistance_cavity),
        XML_Node("RadiationLimitValue", _s.radiation_limit),
        XML_Node("ExcludeWeekends", _s.exclude_weekends),
    ]


# -- VENTILATION --------------------------------------------------------------


def _PhxSpace(_r: spaces.PhxSpace) -> List[xml_writable]:
    return [
        XML_Node("Name", _r.display_name),
        XML_Node("Type", _r.wufi_type),
        XML_Node("IdentNrUtilizationPatternVent", _r.ventilation.schedule.id_num),
        XML_Node("IdentNrVentilationUnit", _r.vent_unit_id_num),
        XML_Node("Quantity", _r.quantity),
        XML_Node("AreaRoom", _r.weighted_floor_area, "unit", "m²"),
        XML_Node("ClearRoomHeight", _r.clear_height, "unit", "m"),
        XML_Node(
            "DesignVolumeFlowRateSupply",
            round(_r.ventilation.load.flow_supply, TOL_LEV1),
            "unit",
            "m³/h",
        ),
        XML_Node(
            "DesignVolumeFlowRateExhaust",
            round(_r.ventilation.load.flow_extract, TOL_LEV1),
            "unit",
            "m³/h",
        ),
        # XML_Node('SupplyFlowRateUserDef', 'Test', "unit", "m³/h"),
        # XML_Node('ExhaustFlowRateUserDef', 'Test', "unit", "m³/h"),
        # XML_Node('DesignFlowInterzonalUserDef', 'Test', "unit", "m³/h"),
    ]


def _UtilizationPatternVent(
    _vent_sched: ventilation.PhxScheduleVentilation,
) -> List[xml_writable]:
    op_periods = _vent_sched.operating_periods
    return [
        XML_Node("Name", _vent_sched.name),
        XML_Node("IdentNr", _vent_sched.id_num),
        XML_Node("OperatingDays", _vent_sched.operating_days),
        XML_Node("OperatingWeeks", _vent_sched.operating_weeks),
        XML_Node("Maximum_DOS", round(op_periods.high.period_operating_hours, TOL_LEV1)),
        XML_Node("Maximum_PDF", round(op_periods.high.period_operation_speed, TOL_LEV1)),
        XML_Node("Standard_DOS", round(op_periods.standard.period_operating_hours, TOL_LEV1)),
        XML_Node("Standard_PDF", round(op_periods.standard.period_operation_speed, TOL_LEV1)),
        XML_Node("Basic_DOS", round(op_periods.basic.period_operating_hours, TOL_LEV1)),
        XML_Node("Basic_PDF", round(op_periods.basic.period_operation_speed, TOL_LEV1)),
        XML_Node("Minimum_DOS", round(op_periods.minimum.period_operating_hours, TOL_LEV1)),
        XML_Node("Minimum_PDF", round(op_periods.minimum.period_operation_speed, TOL_LEV1)),
    ]


# -- MECHANICAL DEVICES -------------------------------------------------------


def _PhxDeviceVentilator(_d: hvac.PhxDeviceVentilator) -> List[xml_writable]:
    return [
        XML_Node("Name", _d.display_name),
        XML_Node("IdentNr", _d.id_num),
        XML_Node("SystemType", _d.system_type.value),
        XML_Node("TypeDevice", _d.device_type.value),
        XML_Node("UsedFor_Heating", _d.usage_profile.space_heating),
        XML_Node("UsedFor_DHW", _d.usage_profile.dhw_heating),
        XML_Node("UsedFor_Cooling", _d.usage_profile.cooling),
        XML_Node("UsedFor_Ventilation", _d.usage_profile.ventilation),
        XML_Node("UsedFor_Humidification", _d.usage_profile.humidification),
        XML_Node("UsedFor_Dehumidification", _d.usage_profile.dehumidification),
        XML_Node("UseOptionalClimate", False),
        XML_Node("IdentNr_OptionalClimate", -1),
        XML_Node("HeatRecovery", _d.params.sensible_heat_recovery),
        XML_Node("MoistureRecovery ", _d.params.latent_heat_recovery),
        XML_Object("PH_Parameters", _d.params, _schema_name="_DeviceVentilatorPhParams"),
    ]


def _DeviceVentilatorPhParams(_p: hvac.PhxDeviceVentilatorParams) -> List[xml_writable]:
    return [
        XML_Node("Quantity", _p.quantity),
        XML_Node("HumidityRecoveryEfficiency", _p.latent_heat_recovery),
        XML_Node("ElectricEfficiency", _p.electric_efficiency),
        XML_Node("DefrostRequired", _p.frost_protection_reqd),
        XML_Node("FrostProtection", _p.frost_protection_reqd),
        XML_Node("TemperatureBelowDefrostUsed", _p.temperature_below_defrost_used),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
        XML_Node("NoSummerBypass", False),
        # XML_Node("SubsoilHeatExchangeEfficiency", _p.),
        # XML_Node("VolumeFlowRateFrom", "unit","m³/h", _p.),
        # XML_Node("VolumeFlowRateTo", "unit","m³/h", _p.),
        # XML_Node("Maximum_VOS", _p.),
        # XML_Node("Maximum_PP", _p.),
        # XML_Node("Standard_VOS", _p.),
        # XML_Node("Standard_PP", _p.),
        # XML_Node("Basic_VOS", _p.),
        # XML_Node("Basic_PP", _p.),
        # XML_Node("Minimum_VOS", _p.),
        # XML_Node("Minimum_PP", _p.),
        # XML_Node("AuxiliaryEnergy", _p.),
        # XML_Node("AuxiliaryEnergyDHW", _p.),
    ]


def _PhxExhaustVentilator(_v: hvac.AnyPhxExhaustVent) -> List[xml_writable]:
    return [
        XML_Node("Name", _v.display_name),
        XML_Node("Type", _v.params.exhaust_type.value),
        XML_Node("ExhaustVolumeFlowRate", _v.params.exhaust_flow_rate_m3h, "unit", "m³/h"),
        XML_Node("RunTimePerYear", _v.params.annual_runtime_minutes),
    ]


def _PhxSupportiveDevice(d: hvac.PhxSupportiveDevice) -> List[xml_writable]:
    return [
        XML_Node("Name", d.display_name),
        XML_Node("Type", d.device_type.value),
        XML_Node("Quantity", d.quantity),
        XML_Node("InConditionedSpace", d.params.in_conditioned_space),
        XML_Node("NormEnergyDemand", d.params.norm_energy_demand_W),
        XML_Node("PeriodOperation", d.params.annual_period_operation_khrs),
    ]


# -- Elec Heating--------------------------------------------------------------


def _PhxDeviceHeaterElec(_d: hvac.PhxHeaterElectric) -> List[xml_writable]:
    return [
        XML_Node("Name", _d.display_name),
        XML_Node("IdentNr", _d.id_num),
        XML_Node("SystemType", _d.system_type.value),
        XML_Node("TypeDevice", _d.device_type.value),
        XML_Node("UsedFor_Heating", _d.usage_profile.space_heating),
        XML_Node("UsedFor_DHW", _d.usage_profile.dhw_heating),
        XML_Node("UsedFor_Cooling", _d.usage_profile.cooling),
        XML_Node("UsedFor_Ventilation", _d.usage_profile.ventilation),
        XML_Node("UsedFor_Humidification", _d.usage_profile.humidification),
        XML_Node("UsedFor_Dehumidification", _d.usage_profile.dehumidification),
        XML_Object("PH_Parameters", _d.params, _schema_name="_DeviceHeaterElecPhParams"),
        XML_Object("DHW_Parameters", _d, _schema_name="_DeviceHeaterElecDeviceParams"),
        XML_Object("Heating_Parameters", _d, _schema_name="_DeviceHeaterHeatingDeviceParams"),
    ]


def _DeviceHeaterElecPhParams(_p: hvac.PhxMechanicalDeviceParams) -> List[xml_writable]:
    return [
        XML_Node("AuxiliaryEnergy", _p.aux_energy),
        XML_Node("AuxiliaryEnergyDHW", _p.aux_energy_dhw),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
    ]


def _DeviceHeaterElecDeviceParams(_d: hvac.PhxHeaterElectric) -> List[xml_writable]:
    _d.usage_profile.space_heating_percent
    return [
        XML_Node("CoverageWithinSystem", _d.usage_profile.dhw_heating_percent),
        XML_Node("Unit", _d.unit),
        XML_Node("Selection", 1),
    ]


def _DeviceHeaterHeatingDeviceParams(_d: hvac.PhxHeaterElectric) -> List[xml_writable]:
    _d.usage_profile.space_heating_percent
    return [
        XML_Node("CoverageWithinSystem", _d.usage_profile.space_heating_percent),
        XML_Node("Unit", _d.unit),
        XML_Node("Selection", 1),
    ]


# -- Boilers ------------------------------------------------------------------


def _PhxDeviceHeaterBoiler(_d: hvac.AnyPhxHeaterBoiler) -> List[xml_writable]:
    ph_params = {
        "NATURAL_GAS": "_DeviceHeaterBoilerFossilPhParams",
        "OIL": "_DeviceHeaterBoilerFossilPhParams",
        "WOOD_LOG": "_DeviceHeaterBoilerWoodPhParams",
        "WOOD_PELLET": "_DeviceHeaterBoilerWoodPhParams",
    }
    return [
        XML_Node("Name", _d.display_name),
        XML_Node("IdentNr", _d.id_num),
        XML_Node("SystemType", _d.system_type.value),
        XML_Node("TypeDevice", _d.device_type.value),
        XML_Node("UsedFor_Heating", _d.usage_profile.space_heating),
        XML_Node("UsedFor_DHW", _d.usage_profile.dhw_heating),
        XML_Node("UsedFor_Cooling", _d.usage_profile.cooling),
        XML_Node("UsedFor_Ventilation", _d.usage_profile.ventilation),
        XML_Node("UsedFor_Humidification", _d.usage_profile.humidification),
        XML_Node("UsedFor_Dehumidification", _d.usage_profile.dehumidification),
        XML_Object("PH_Parameters", _d.params, _schema_name=ph_params[_d.params.fuel.name]),
        XML_Object("DHW_Parameters", _d, _schema_name="_DeviceHeaterBoilerDHWParams"),
        XML_Object("Heating_Parameters", _d, _schema_name="_DeviceHeaterBoilerHeatingParams"),
    ]


def _DeviceHeaterBoilerWoodPhParams(
    _p: hvac.PhxHeaterBoilerWoodParams,
) -> List[xml_writable]:
    return [
        XML_Node("EnergySourceBoilerType", _p.fuel.value),
        XML_Node("MaximalBoilerPower", _p.rated_capacity),
        XML_Node("SolarFractionBoilerSpaceHeating", _p.solar_fraction),
        XML_Node("EfficiencyHeatGeneratorBasicCycle", _p.effic_in_basic_cycle),
        XML_Node("EfficiencyHeatGeneratorConstantOperation", _p.effic_in_const_operation),
        XML_Node("AverageFractionHeatOutputReleasedHeatingCircuit", _p.avg_frac_heat_output),
        XML_Node("TemperatureDifferencePowerOnPowerOff", _p.temp_diff_on_off),
        XML_Node("UsefulHeatOutputBasicCycl", _p.rated_capacity),
        XML_Node("AveragePowerOutputHeatGenerator", _p.rated_capacity),
        XML_Node("DemandBasicCycle", _p.demand_basic_cycle),
        XML_Node("PowerConsumptionStationarRun", _p.power_standard_run),
        XML_Node("NoTransportPellets", _p.no_transport_pellets),
        XML_Node("OnlyControl", _p.only_control),
        XML_Node("AreaMechanicalRoom", _p.area_mech_room),
        XML_Node("AuxiliaryEnergy", _p.aux_energy),
        XML_Node("AuxiliaryEnergyDHW", _p.aux_energy_dhw),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
    ]


def _DeviceHeaterBoilerFossilPhParams(
    _p: hvac.PhxHeaterBoilerFossilParams,
) -> List[xml_writable]:
    return [
        XML_Node("EnergySourceBoilerType", _p.fuel.value),
        XML_Node("CondensingBoiler", _p.condensing),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
        XML_Node("MaximalBoilerPower", _p.rated_capacity),
        XML_Node("BoilerEfficiency30", _p.effic_at_30_percent_load),
        XML_Node("BoilerEfficiencyNominalOutput", _p.effic_at_nominal_load),
        XML_Node("AverageReturnTemperatureMeasured30Load", _p.avg_rtrn_temp_at_30_percent_load),
        XML_Node("AverageBoilerTemperatureDesign70_55", _p.avg_temp_at_70C_55C),
        XML_Node("AverageBoilerTemperatureDesign55_45", _p.avg_temp_at_55C_45C),
        XML_Node("AverageBoilerTemperatureDesign35_28", _p.avg_temp_at_32C_28C),
        XML_Node("StandbyHeatLossBoiler70", _p.standby_loss_at_70C),
        XML_Node("SolarFractionBoilerSpaceHeating", _p.aux_energy),
        XML_Node("AuxiliaryEnergy", _p.aux_energy),
        XML_Node("AuxiliaryEnergyDHW", _p.aux_energy_dhw),
    ]


def _DeviceHeaterBoilerDHWParams(_d: hvac.AnyPhxHeaterBoiler) -> List[xml_writable]:
    return [
        XML_Node("CoverageWithinSystem", _d.usage_profile.dhw_heating_percent),
        XML_Node("Unit", _d.unit),
        XML_Node("Selection", 1),
    ]


def _DeviceHeaterBoilerHeatingParams(_d: hvac.AnyPhxHeaterBoiler) -> List[xml_writable]:
    return [
        XML_Node("CoverageWithinSystem", _d.usage_profile.space_heating_percent),
        XML_Node("Unit", _d.unit),
        XML_Node("Selection", 1),
    ]


# -- District Heat ------------------------------------------------------------


def _PhxDeviceHeaterDistrict(_d: hvac.PhxHeaterDistrictHeat) -> List[xml_writable]:
    return [
        XML_Node("Name", _d.display_name),
        XML_Node("IdentNr", _d.id_num),
        XML_Node("SystemType", _d.system_type.value),
        XML_Node("TypeDevice", _d.device_type.value),
        XML_Node("UsedFor_Heating", _d.usage_profile.space_heating),
        XML_Node("UsedFor_DHW", _d.usage_profile.dhw_heating),
        XML_Node("UsedFor_Cooling", _d.usage_profile.cooling),
        XML_Node("UsedFor_Ventilation", _d.usage_profile.ventilation),
        XML_Node("UsedFor_Humidification", _d.usage_profile.humidification),
        XML_Node("UsedFor_Dehumidification", _d.usage_profile.dehumidification),
        XML_Object("DHW_Parameters", _d, _schema_name="_DeviceHeaterDistrictDeviceParams"),
        XML_Object("Heating_Parameters", _d, _schema_name="_DeviceHeaterDistrictDeviceParams"),
    ]


def _DeviceHeaterDistrictDeviceParams(
    _d: hvac.PhxHeaterDistrictHeat,
) -> List[xml_writable]:
    return [
        XML_Node("CoverageWithinSystem", _d.percent_coverage),
        XML_Node("Unit", _d.unit),
        XML_Node("Selection", 1),
    ]


# -- Heat Pumps ---------------------------------------------------------------


def _PhxDeviceHeaterHeatPump(_d: hvac.PhxHeatPumpDevice) -> List[xml_writable]:
    param_schemas: Dict[hvac.HeatPumpType, str] = {
        hvac.PhxHeatPumpAnnualParams.hp_type: "_DeviceHeaterHeatPumpPhParamsAnnual",
        hvac.PhxHeatPumpMonthlyParams.hp_type: "_DeviceHeaterHeatPumpPhParamsMonthly",
        hvac.PhxHeatPumpHotWaterParams.hp_type: "_DeviceHeaterHeatPumpPhParamsHotWater",
        hvac.PhxHeatPumpCombinedParams.hp_type: "_DeviceHeaterHeatPumpPhParamsCombined",
    }

    hp_type: Optional[hvac.HeatPumpType] = getattr(_d.params, "hp_type", None)
    if not hp_type:
        return []

    return [
        XML_Node("Name", _d.display_name),
        XML_Node("IdentNr", _d.id_num),
        XML_Node("SystemType", _d.system_type.value),
        XML_Node("TypeDevice", _d.device_type.value),
        XML_Node("UsedFor_Heating", _d.usage_profile.space_heating),
        XML_Node("UsedFor_DHW", _d.usage_profile.dhw_heating),
        XML_Node("UsedFor_Cooling", _d.usage_profile.cooling),
        XML_Node("UsedFor_Ventilation", _d.usage_profile.ventilation),
        XML_Node("UsedFor_Humidification", _d.usage_profile.humidification),
        XML_Node("UsedFor_Dehumidification", _d.usage_profile.dehumidification),
        XML_Object("PH_Parameters", _d.params, _schema_name=param_schemas[hp_type]),
        XML_Object("DHW_Parameters", _d, _schema_name="_DeviceHotWaterHeatPumpDeviceParams"),
        XML_Object("Heating_Parameters", _d, _schema_name="_DeviceHeaterHeatPumpDeviceParams"),
        XML_Object("Cooling_Parameters", _d, _schema_name="_DeviceCoolingHeatPumpDeviceParams"),
    ]


def _DeviceHeaterHeatPumpPhParamsAnnual(
    _p: hvac.PhxHeatPumpAnnualParams,
) -> List[xml_writable]:
    return [
        XML_Node("AuxiliaryEnergy", _p.aux_energy),
        XML_Node("AuxiliaryEnergyDHW", _p.aux_energy_dhw),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
        XML_Node("AnnualCOP", _p.annual_COP),
        XML_Node("TotalSystemPerformanceRatioHeatGenerator", _p.total_system_perf_ratio),
        XML_Node("HPType", _p.hp_type.value),
    ]


def _DeviceHeaterHeatPumpPhParamsMonthly(
    _p: hvac.PhxHeatPumpMonthlyParams,
) -> List[xml_writable]:
    return [
        XML_Node("AuxiliaryEnergy", _p.aux_energy),
        XML_Node("AuxiliaryEnergyDHW", _p.aux_energy_dhw),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
        XML_Node("RatedCOP1", _p.COP_1),
        XML_Node("RatedCOP2", _p.COP_2),
        XML_Node("AmbientTemperature1", _p.ambient_temp_1),
        XML_Node("AmbientTemperature2", _p.ambient_temp_2),
        XML_Node("HPType", _p.hp_type.value),
    ]


def _DeviceHeaterHeatPumpPhParamsHotWater(
    _p: hvac.PhxHeatPumpHotWaterParams,
) -> List[xml_writable]:
    return [
        XML_Node("AuxiliaryEnergy", _p.aux_energy),
        XML_Node("AuxiliaryEnergyDHW", _p.aux_energy_dhw),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
        XML_Node("AnnualCOP", _p.annual_COP),
        XML_Node("TotalSystemPerformanceRatioHeatGenerator", _p.total_system_perf_ratio),
        XML_Node("HPWH_EF", _p.annual_energy_factor),
        XML_Node("HPType", _p.hp_type.value),
    ]


def _DeviceHeaterHeatPumpPhParamsCombined(
    _p: hvac.PhxHeatPumpCombinedParams,
) -> List[xml_writable]:
    return [
        XML_Node("AuxiliaryEnergy", _p.aux_energy),
        XML_Node("AuxiliaryEnergyDHW", _p.aux_energy_dhw),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
        XML_Node("HPType", _p.hp_type.value),
    ]


def _DeviceHotWaterHeatPumpDeviceParams(_d: hvac.PhxHeatPumpDevice) -> List[xml_writable]:
    return [
        XML_Node("CoverageWithinSystem", _d.usage_profile.dhw_heating_percent),
        XML_Node("Unit", _d.unit),
        XML_Node("Selection", 1),
    ]


def _DeviceHeaterHeatPumpDeviceParams(_d: hvac.PhxHeatPumpDevice) -> List[xml_writable]:
    return [
        XML_Node("CoverageWithinSystem", _d.usage_profile.space_heating_percent),
        XML_Node("Unit", _d.unit),
        XML_Node("Selection", 1),
    ]


def _DeviceCoolingHeatPumpDeviceParams(_d: hvac.PhxHeatPumpDevice) -> List[xml_writable]:
    return [
        XML_Node("CoverageWithinSystem", _d.usage_profile.cooling_percent),
        XML_Node("Unit", _d.unit),
        XML_Node("Selection", 1),
    ]


def _PhxDeviceWaterStorage(_d: hvac.PhxHotWaterTank) -> List[xml_writable]:
    return [
        XML_Node("Name", _d.display_name),
        XML_Node("IdentNr", _d.id_num),
        XML_Node("SystemType", _d.system_type.value),
        XML_Node("TypeDevice", _d.device_type.value),
        XML_Node("UsedFor_Heating", _d.usage_profile.space_heating),
        XML_Node("UsedFor_DHW", _d.usage_profile.dhw_heating),
        XML_Node("UsedFor_Cooling", _d.usage_profile.cooling),
        XML_Node("UsedFor_Ventilation", _d.usage_profile.ventilation),
        XML_Node("UsedFor_Humidification", _d.usage_profile.humidification),
        XML_Node("UsedFor_Dehumidification", _d.usage_profile.dehumidification),
        XML_Object("PH_Parameters", _d, _schema_name="_DeviceWaterStoragePhParams"),
    ]


def _DeviceWaterStoragePhParams(_t: hvac.PhxHotWaterTank) -> List[xml_writable]:
    return [
        XML_Node("SolarThermalStorageCapacity", _t.params.storage_capacity),
        XML_Node("StorageLossesStandby", _t.params.standby_losses),
        XML_Node("TotalSolarThermalStorageLosses", _t.params.standby_losses),
        XML_Node("InputOption", _t.params.input_option.value),
        XML_Node("AverageHeatReleaseStorage", _t.params.storage_loss_rate),
        XML_Node("TankRoomTemp ", _t.params.room_temp),
        XML_Node("TypicalStorageWaterTemperature", _t.params.water_temp),
        XML_Node("QauntityWS", _t.quantity),
        XML_Node("AuxiliaryEnergy", _t.params.aux_energy),
        XML_Node("AuxiliaryEnergyDHW", _t.params.aux_energy_dhw),
        XML_Node("InConditionedSpace", _t.params.in_conditioned_space),
    ]


# -- MECHANICAL SYSTEMS / DISTRIBUTION ----------------------------------------


def _DistributionDHWTwig(_twg: hvac.PhxPipeElement) -> List[xml_writable]:
    diameter_inches = convert(_twg.weighted_diameter_mm, "MM", "IN") or 0.0
    diameter_type = PhxHotWaterPipingInchDiameterType.nearest_key(diameter_inches)
    return [
        XML_Node("Name", _twg.display_name),
        XML_Node("IdentNr", _twg.id_num),
        XML_Node("PipingLength", _twg.length_m),
        XML_Node("PipeMaterial", _twg.material.value),
        XML_Node("PipingDiameter", diameter_type.value),
    ]


def _DistributionDHWBranch(_br: hvac.PhxPipeBranch) -> List[xml_writable]:
    diameter_inches = convert(_br.pipe_element.weighted_diameter_mm, "MM", "IN") or 0.0
    diameter_type = PhxHotWaterPipingInchDiameterType.nearest_key(diameter_inches)
    return [
        XML_Node("Name", _br.display_name),
        XML_Node("IdentNr", _br.id_num),
        XML_Node("PipingLength", _br.pipe_element.length_m),
        XML_Node("PipeMaterial", _br.pipe_element.material.value),
        XML_Node("PipingDiameter", diameter_type.value),
        XML_List(
            "Twigs",
            [
                XML_Object("Twig", br, "index", i, _schema_name="_DistributionDHWTwig")
                for i, br in enumerate(_br.fixtures)
            ],
        ),
    ]


def _DistributionDHWTrunc(_t: hvac.PhxPipeTrunk) -> List[xml_writable]:
    diameter_inches = convert(_t.pipe_element.weighted_diameter_mm, "MM", "IN") or 0.0
    diameter_type = PhxHotWaterPipingInchDiameterType.nearest_key(diameter_inches)
    return [
        XML_Node("Name", _t.display_name),
        XML_Node("IdentNr", _t.id_num),
        XML_Node("PipingLength", _t.pipe_element.length_m),
        XML_Node("PipeMaterial", _t.pipe_element.material.value),
        XML_Node("PipingDiameter", diameter_type.value),
        XML_Node("CountUnitsOrFloors", _t.multiplier),
        XML_Node("DemandRecirculation", _t.pipe_element.demand_recirculation),
        XML_List(
            "Branches",
            [
                XML_Object("Branch", br, "index", i, _schema_name="_DistributionDHWBranch")
                for i, br in enumerate(_t.branches)
            ],
        ),
    ]


def _DistributionDHW(_c: hvac.PhxMechanicalSystemCollection):
    # -- alias
    p = _c._distribution_hw_recirculation_params
    return [
        # -- Settings
        XML_Node("CalculationMethodIndividualPipes", p.calc_method.value),
        XML_Node("DemandRecirculation", p.demand_recirc),
        XML_Node("SelectionhotWaterFixtureEff", p.hot_water_fixtures),
        XML_Node("NumberOfBathrooms", p.num_bathrooms),
        XML_Node("AllPipesAreInsulated", p.all_pipes_insulated),
        XML_Node("SelectionUnitsOrFloors", p.units_or_floors.value),
        XML_Node("PipeMaterialSimplifiedMethod", p.pipe_material.value),
        XML_Node("PipeDiameterSimplifiedMethod", p.pipe_diameter),
        # -- Simplified Method
        XML_Node("TemperatureRoom_WR", p.air_temp),
        XML_Node("DesignFlowTemperature_WR", p.water_temp),
        XML_Node("DailyRunningHoursCirculation_WR", p.daily_recirc_hours),
        XML_Node("LengthCirculationPipes_WR", _c.dhw_recirc_total_length_m),
        XML_Node(
            "HeatLossCoefficient_WR",
            _c.dhw_recirc_weighted_heat_loss_coeff,
        ),
        XML_Node("LengthIndividualPipes_WR", _c.dhw_distribution_total_length_m),
        XML_Node(
            "ExteriorPipeDiameter_WR",
            _c.dhw_distribution_weighted_diameter_mm,
        ),
        # -- Detailed Trunk, Branch, Fixture (Twig) piping
        XML_List(
            "Truncs",
            [
                XML_Object("Trunc", trunc, "index", i, _schema_name="_DistributionDHWTrunc")
                for i, trunc in enumerate(_c.dhw_distribution_trunks)
            ],
        ),
    ]


class DistributionHeating:
    def __init__(self) -> None:
        raise NotImplementedError


def _DistributionHeating(_d):
    raise NotImplementedError
    # return [
    #     XML_Node("LengthPipes_WR", _d.),
    #     XML_Node("LengthPipes_CR1", _d.),
    #     XML_Node("LengthPipes_CR2", _d.),
    #     XML_Node("HeatLossCoefficient_WR", _d.),
    #     XML_Node("HeatLossCoefficient_CR1", _d.),
    #     XML_Node("HeatLossCoefficient_CR2", _d.),
    #     XML_Node("TemperatureRoom_WR", _d.),
    #     XML_Node("TemperatureRoom_CR1", _d.),
    #     XML_Node("TemperatureRoom_CR2", _d.),
    #     XML_Node("DesignFlowTemperature_WR", _d.),
    #     XML_Node("DesignFlowTemperature_CR1", _d.),
    #     XML_Node("DesignFlowTemperature_CR2", _d.),
    #     XML_Node("DesignSystemHeatingLoad_WR", _d.),
    #     XML_Node("DesignSystemHeatingLoad_CR1", _d.),
    #     XML_Node("DesignSystemHeatingLoad_CR2", _d.),
    #     XML_Node("FlowTControl_WR", _d.),
    #     XML_Node("FlowTControl_CR1", _d.),
    #     XML_Node("FlowTControl_CR2", _d.),
    # ]


def _PhxDuctElement(_d: hvac.PhxDuctElement) -> List[xml_writable]:
    return [
        XML_Node("Name", _d.display_name),
        XML_Node("IdentNr", _d.id_num),
        XML_Node("DuctDiameter", _d.diameter_mm, "unit", "mm"),
        XML_Node("DuctShapeHeight", _d.height_mm, "unit", "mm"),
        XML_Node("DuctShapeWidth", _d.width_mm, "unit", "mm"),
        XML_Node("DuctLength", _d.length_m, "unit", "m"),
        XML_Node("InsulationThickness", _d.insulation_thickness_mm, "unit", "mm"),
        XML_Node("ThermalConductivity", _d.insulation_conductivity_wmk, "unit", "W/mK"),
        XML_Node("Quantity", _d.quantity, "unit", "-"),
        XML_Node("DuctType", _d.duct_type.value),
        XML_Node("DuctShape", _d.duct_shape),
        XML_Node("IsReflective", _d.is_reflective),
        XML_List(
            "AssignedVentUnits",
            [XML_Node("IdentNrVentUnit", id, "index", i) for i, id in enumerate(_d.assigned_vent_unit_ids)],
        ),
    ]


# -- MECHANICAL SYSTEMS / COOLING-DISTRIBUTION --------------------------------
# -- DEV-NOTE: I don't want to have the Cooling 'distribution' as part of the PHX model
# -- separate from the actual mechanical devices. It is stupid that things like COP are
# -- stored in there. So use a temp class to unwrap all the relevant data from the devices.


T = TypeVar("T", bound=hvac.AnyPhxCoolingParamsType)


class TempDistributionCooling:
    """Temporary wrapper class for WUFI format Cooling Distribution data."""

    def __init__(self, _devices: List[hvac.PhxHeatPumpDevice]) -> None:
        # -- If there are no heat-pump systems, the input list will be empty.
        # -- Have to combine the heat-pump parameters all together for each type.
        self.ventilation_params = self.sum_params([d.params_cooling.ventilation for d in _devices])
        self.recirculation_params = self.sum_params([d.params_cooling.recirculation for d in _devices])
        self.dehumidification_params = self.sum_params([d.params_cooling.dehumidification for d in _devices])
        self.panel_params = self.sum_params([d.params_cooling.panel for d in _devices])

    def sum_params(self, _cooling_params: List[T]) -> Optional[T]:
        """Returns a single HVAC Cooling Params, made from a list of input devices, or None if no devices input."""
        if not _cooling_params:
            return None

        return reduce(lambda a, b: a + b, _cooling_params, _cooling_params[0])

    def __bool__(self) -> bool:
        return all(
            [
                self.ventilation_params is not None,
                self.recirculation_params is not None,
                self.dehumidification_params is not None,
                self.panel_params is not None,
            ]
        )


def _DistributionCooling(_clg_distr: TempDistributionCooling) -> List[xml_writable]:
    base = []

    # -- If its a heating system, will not have any cooling params.
    if not _clg_distr:
        return base

    if _clg_distr.ventilation_params and _clg_distr.ventilation_params.used:
        vent_params: hvac.PhxCoolingVentilationParams = _clg_distr.ventilation_params
        base += [
            XML_Node("CoolingViaVentilationAir", True),
            XML_Node("SupplyAirCoolingOnOff", vent_params.single_speed),
            XML_Node("MaxSupplyAirCoolingPower", vent_params.capacity),
            XML_Node("MinTemperatureCoolingCoilSupplyAir", vent_params.min_coil_temp),
            XML_Node("SupplyAirCoolinCOP", vent_params.annual_COP),
        ]
    if _clg_distr.recirculation_params and _clg_distr.recirculation_params.used:
        recirc_params: hvac.PhxCoolingRecirculationParams = _clg_distr.recirculation_params
        base += [
            XML_Node("CoolingViaRecirculation", True),
            XML_Node("RecirculatingAirOnOff", recirc_params.single_speed),
            XML_Node("MaxRecirculationAirCoolingPower", recirc_params.capacity),
            XML_Node("MinTempCoolingCoilRecirculatingAir", recirc_params.min_coil_temp),
            XML_Node("RecirculationCoolingCOP", recirc_params.annual_COP),
            XML_Node("RecirculationAirVolume", recirc_params.flow_rate_m3_hr),
            XML_Node("ControlledRecirculationVolumeFlow", recirc_params.flow_rate_variable),
        ]
    if _clg_distr.dehumidification_params and _clg_distr.dehumidification_params.used:
        dehumid_params: hvac.PhxCoolingDehumidificationParams = _clg_distr.dehumidification_params
        base += [
            XML_Node("Dehumidification", True),
            XML_Node("UsefullDehumidificationHeatLoss", dehumid_params.useful_heat_loss),
            XML_Node("DehumdificationCOP", dehumid_params.annual_COP),
            XML_Node("SEER", None),
            XML_Node("EER", None),
            XML_Node("DehumidificationElEnergy", None),
        ]
    if _clg_distr.panel_params and _clg_distr.panel_params.used:
        panel_params: hvac.PhxCoolingPanelParams = _clg_distr.panel_params
        base += [
            XML_Node("PanelCooling", True),
            XML_Node("DehumdificationCOP", panel_params.annual_COP),
        ]

    return base


def _PHDistribution(_c: hvac.PhxMechanicalSystemCollection):
    return [
        XML_Object("DistributionDHW", _c, _schema_name="_DistributionDHW"),
        # XML_Object('DistributionHeating', DistributionHeating()),
        XML_Object(
            "DistributionCooling",
            TempDistributionCooling(_c.heat_pump_devices),
            _schema_name="_DistributionCooling",
        ),
        XML_List(
            "DistributionVentilation",
            [XML_Object("Duct", d, "index", i) for i, d in enumerate(_c.vent_ducting)],
        ),
        XML_Node("UseDefaultValues", False),
        XML_Node("DeviceInConditionedSpace", True),
        XML_List(
            "SupportiveDevices",
            [XML_Object("SupportiveDevice", d, "index", i) for i, d in enumerate(_c.supportive_devices)],
        ),
    ]


# -- MECHANICAL SYSTEMS / COLLECTIONS -----------------------------------------


def _PhxZoneCoverage(_zc: hvac.PhxZoneCoverage) -> List[xml_writable]:
    return [
        XML_Node("IdentNrZone", int(_zc.zone_num)),
        XML_Node("CoverageHeating", _zc.heating),
        XML_Node("CoverageCooling", _zc.cooling),
        XML_Node("CoverageVentilation", _zc.ventilation),
        XML_Node("CoverageHumidification", _zc.humidification),
        XML_Node("CoverageDehumidification", _zc.dehumidification),
    ]


def _PhxMechanicalSystemCollection(
    _hvac_collection: hvac.PhxMechanicalSystemCollection,
) -> List[xml_writable]:
    devices = {
        hvac.DeviceType.VENTILATION: "_PhxDeviceVentilator",
        hvac.DeviceType.ELECTRIC: "_PhxDeviceHeaterElec",
        hvac.DeviceType.BOILER: "_PhxDeviceHeaterBoiler",
        hvac.DeviceType.DISTRICT_HEAT: "_PhxDeviceHeaterDistrict",
        hvac.DeviceType.HEAT_PUMP: "_PhxDeviceHeaterHeatPump",
        hvac.DeviceType.WATER_STORAGE: "_PhxDeviceWaterStorage",
        hvac.DeviceType.PHOTOVOLTAIC: "_PhxDevicePhotovoltaic",
    }

    wufi_devices: List[hvac.AnyMechDevice] = []
    wufi_devices.extend(_hvac_collection.devices)
    wufi_devices.extend(_hvac_collection.renewable_devices)
    return [
        XML_Node("Name", _hvac_collection.display_name),
        XML_Node("Type", _hvac_collection.sys_type_num, "choice", _hvac_collection.sys_type_str),
        XML_Node("IdentNr", _hvac_collection.id_num),
        XML_List(
            "ZonesCoverage",
            [XML_Object("ZoneCoverage", n, "index", i) for i, n in enumerate([_hvac_collection.zone_coverage])],
        ),
        XML_List(
            "Devices",
            [
                XML_Object("Device", d, "index", i, _schema_name=devices[d.device_type])
                for i, d in enumerate(wufi_devices)
            ],
        ),
        XML_Object("PHDistribution", _hvac_collection, _schema_name="_PHDistribution"),
    ]


# -- ELEC. EQUIPMENT DEVICES --------------------------------------------------


def _PhxDeviceDishwasher(_d: elec_equip.PhxDeviceDishwasher) -> List[xml_writable]:
    return [
        XML_Node("Type", _d.device_type.value),
        XML_Node("Connection", _d.water_connection),
        XML_Node("DishwasherCapacityPreselection", _d.capacity_type),
        XML_Node("DishwasherCapacityInPlace", _d.capacity),
    ]


def _PhxDeviceClothesWasher(_d: elec_equip.PhxDeviceClothesWasher) -> List[xml_writable]:
    return [
        XML_Node("Type", _d.device_type.value),
        XML_Node("Connection", _d.water_connection),
        XML_Node("UtilizationFactor", _d.utilization_factor),
        XML_Node("CapacityClothesWasher", _d.capacity),
        XML_Node("MEF_ModifiedEnergyFactor", _d.modified_energy_factor),
    ]


def _PhxDeviceClothesDryer(_d: elec_equip.PhxDeviceClothesDryer) -> List[xml_writable]:
    return [
        XML_Node("Type", _d.device_type.value),
        XML_Node("Dryer_Choice", _d.dryer_type),
        XML_Node("GasConsumption", _d.gas_consumption),
        XML_Node("EfficiencyFactorGas", _d.gas_efficiency_factor),
        XML_Node("FieldUtilizationFactorPreselection", _d.field_utilization_factor_type),
        XML_Node("FieldUtilizationFactor", _d.field_utilization_factor),
    ]


def _PhxDeviceRefrigerator(_d: elec_equip.PhxDeviceRefrigerator) -> List[xml_writable]:
    return [
        XML_Node("Type", _d.device_type.value),
    ]


def _PhxDeviceFreezer(_d: elec_equip.PhxDeviceFreezer) -> List[xml_writable]:
    return [
        XML_Node("Type", 5),
    ]


def _PhxDeviceFridgeFreezer(_d: elec_equip.PhxDeviceFridgeFreezer) -> List[xml_writable]:
    return [
        XML_Node("Type", _d.device_type.value),
    ]


def _PhxDeviceCooktop(_d: elec_equip.PhxDeviceCooktop) -> List[xml_writable]:
    return [
        XML_Node("Type", _d.device_type.value),
        XML_Node("CookingWith", _d.cooktop_type),
    ]


def _PhxDeviceMEL(_d: elec_equip.PhxDeviceMEL) -> List[xml_writable]:
    return [
        XML_Node("Type", _d.device_type.value),
    ]


def _PhxDeviceLightingInterior(
    _d: elec_equip.PhxDeviceLightingInterior,
) -> List[xml_writable]:
    return [
        XML_Node("Type", _d.device_type.value),
        XML_Node("FractionHightEfficiency", _d.frac_high_efficiency),
    ]


def _PhxDeviceLightingExterior(
    _d: elec_equip.PhxDeviceLightingExterior,
) -> List[xml_writable]:
    return [
        XML_Node("Type", _d.device_type.value),
        XML_Node("FractionHightEfficiency", _d.frac_high_efficiency),
    ]


def _PhxDeviceLightingGarage(
    _d: elec_equip.PhxDeviceLightingGarage,
) -> List[xml_writable]:
    return [
        XML_Node("Type", _d.device_type.value),
        XML_Node("FractionHightEfficiency", _d.frac_high_efficiency),
    ]


def _PhxDeviceCustomElec(_d: elec_equip.PhxDeviceCustomElec) -> List[xml_writable]:
    return [
        XML_Node("Type", _d.device_type.value),
    ]


def _PhxDeviceCustomLighting(
    _d: elec_equip.PhxDeviceCustomLighting,
) -> List[xml_writable]:
    return [
        XML_Node("Type", _d.device_type.value),
    ]


def _PhxDeviceCustomMEL(_d: elec_equip.PhxDeviceCustomMEL) -> List[xml_writable]:
    return [
        XML_Node("Type", _d.device_type.value),
    ]


def _PhxElevatorHydraulic(_d: elec_equip.PhxElevatorHydraulic) -> List[xml_writable]:
    return [
        XML_Node("Type", _d.device_type.value),
    ]


def _PhxElevatorGearedTraction(
    _d: elec_equip.PhxElevatorGearedTraction,
) -> List[xml_writable]:
    return [
        XML_Node("Type", _d.device_type.value),
    ]


def _PhxElevatorGearlessTraction(
    _d: elec_equip.PhxElevatorGearlessTraction,
) -> List[xml_writable]:
    return [
        XML_Node("Type", _d.device_type.value),
    ]


def _PhxElectricalDevice(_d: elec_equip.PhxElectricalDevice) -> List[xml_writable]:
    common_attributes = [
        XML_Node("Comment", _d.comment),
        XML_Node("ReferenceQuantity", _d.reference_quantity),
        XML_Node("Quantity", _d.get_quantity()),
        XML_Node("InConditionedSpace", _d.in_conditioned_space),
        XML_Node("ReferenceEnergyDemandNorm", _d.reference_energy_norm),
        XML_Node("EnergyDemandNorm", _d.get_energy_demand()),
        XML_Node("EnergyDemandNormUse", _d.energy_demand_per_use),
        XML_Node("CEF_CombinedEnergyFactor", _d.combined_energy_factor),
    ]

    device_schema = getattr(sys.modules[__name__], f"_{_d.__class__.__name__}")
    appliance_specific_attributes = device_schema(_d)
    return common_attributes + appliance_specific_attributes


# -- NON-RES PROGRAM DATA -----------------------------------------------------


def _UtilizationPattern(_p: occupancy.PhxScheduleOccupancy) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _p.id_num),
        XML_Node("Name", _p.display_name),
        XML_Node("HeightUtilizationLevel", 0.5),
        XML_Node("BeginUtilization", _p.start_hour),
        XML_Node("EndUtilization", _p.end_hour),
        XML_Node("AnnualUtilizationDays", round(_p.annual_utilization_days, TOL_LEV1)),
        XML_Node("IlluminationLevel", 300, "unit", "lux"),
        XML_Node("RelativeAbsenteeism", _p.relative_utilization_factor, "unit", "-"),
        XML_Node("PartUseFactorPeriodForLighting", 1, "unit", "-"),
        # XML_Node("AverageOccupancy", "", "unit", "m²/Person"),
        # XML_Node("RoomSetpointTemperature", "", "unit", "°C"),
        # XML_Node("HeatingTemperatureReduction", "", "unit", "°C"),
        # XML_Node("DailyUtilizationHours", "", "unit", "hrs/d"),
        # XML_Node("AnnualUtilizationHours", "", "unit", "hrs/yr"),
        # XML_Node("AnnualUtilizationHoursDaytime", "", "unit", "hrs/yr"),
        # XML_Node("AnnualUtilizationHoursNighttime", "", "unit", "hrs/yr"),
        # XML_Node("DailyHeatingOperationHours", "", "unit", "hrs/d"),
        # XML_Node("DailyVentilationOperatingHours", "", "unit", "hrs/d"),
        # XML_Node("NumberOfMaxTabsPerDay", "", "unit", "-"),
    ]


def _LoadPerson(_sp: spaces.PhxSpace) -> List[xml_writable]:
    return [
        XML_Node("Name", _sp.display_name),
        # XML_Node("IdentNrUtilizationPattern", 1), #<---- changed
        XML_Node("IdentNrUtilizationPattern", _sp.occupancy.schedule.id_num),  # <---- changed
        XML_Node("ChoiceActivityPersons", 3),
        XML_Node("NumberOccupants", _sp.peak_occupancy),
        XML_Node("FloorAreaUtilizationZone", _sp.floor_area, "unit", "m²"),
    ]


def _LoadsLighting(_sp: spaces.PhxSpace) -> List[xml_writable]:
    return [
        XML_Node("Name", _sp.display_name),
        # -- NOTE: As per discussions with Phius we use the
        # -- occupancy pattern, and then just set the full-load hours directly
        XML_Node("RoomCategory", _sp.occupancy.schedule.id_num),
        XML_Node("ChoiceLightTransmissionGlazing", 1),
        XML_Node("LightingControl", 1),
        XML_Node("WithinThermalEnvelope", True),
        XML_Node("MotionDetector", False),
        XML_Node("FacadeIncludingWindows", False),
        XML_Node("FractionTreatedFloorArea", 1),
        XML_Node("DeviationFromNorth", 0),
        XML_Node("RoomDepth", 1),
        XML_Node("RoomWidth", 1),
        XML_Node("RoomHeight", 1),
        XML_Node("LintelHeight", 1),
        XML_Node("WindowWidth", 1),
        XML_Node("InstalledLightingPower", _sp.lighting.load.installed_w_per_m2),
        XML_Node(
            "LightingFullLoadHours",
            round(_sp.lighting.schedule.full_load_lighting_hours, 1),
        ),
    ]


# -- PV System ----------------------------------------------------------------


def _PhxDevicePhotovoltaicParams(
    _p: renewable_devices.PhxDevicePhotovoltaicParams,
) -> List[xml_writable]:
    return [
        XML_Node("SelectionLocation", _p.location_type),
        XML_Node("SelectionOnSiteUtilization", _p.onsite_utilization_type),
        XML_Node("SelectionUtilization", _p.utilization_type),
        XML_Node("ArraySizePV", _p.array_size),
        XML_Node("PhotovoltaicRenewableEnergy", _p.photovoltaic_renewable_energy),
        XML_Node("OnsiteUtilization", _p.onsite_utilization_factor),
        XML_Node("AuxiliaryEnergy", _p.auxiliary_energy),
        XML_Node("AuxiliaryEnergyDHW", _p.auxiliary_energy_DHW),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
    ]


def _PhxDevicePhotovoltaic(
    _d: renewable_devices.PhxDevicePhotovoltaic,
) -> List[xml_writable]:
    use: _base.PhxUsageProfile = _d.usage_profile
    return [
        XML_Node("Name", _d.display_name),
        XML_Node("IdentNr", _d.id_num),
        XML_Node("SystemType", _d.system_type.value),
        XML_Node("TypeDevice", _d.device_type.value),
        XML_Node("UsedFor_Heating", use.space_heating),
        XML_Node("UsedFor_DHW", use.dhw_heating),
        XML_Node("UsedFor_Cooling", use.cooling),
        XML_Node("UsedFor_Ventilation", use.ventilation),
        XML_Node("UsedFor_Humidification", use.humidification),
        XML_Node("UsedFor_Dehumidification", use.dehumidification),
        XML_Object(
            "PH_Parameters",
            _d.params,
            _schema_name="_PhxDevicePhotovoltaicParams",
        ),
    ]
